# Generated by Django 5.1.4 on 2024-12-21 22:26

import core.models
import django.contrib.auth.validators
import django.db.models.deletion
import django.utils.timezone
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="FeatureFlags",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(editable=False, max_length=100, unique=True)),
                (
                    "description",
                    models.TextField(blank=True, editable=False, max_length=500, null=True),
                ),
                ("value", models.BooleanField(default=False)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Feature Flag",
                "verbose_name_plural": "Feature Flags",
            },
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(blank=True, null=True, verbose_name="last login"),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={"unique": "A user with that username already exists."},
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[django.contrib.auth.validators.UnicodeUsernameValidator()],
                        verbose_name="username",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(blank=True, max_length=150, verbose_name="first name"),
                ),
                (
                    "last_name",
                    models.CharField(blank=True, max_length=150, verbose_name="last name"),
                ),
                (
                    "email",
                    models.EmailField(blank=True, max_length=254, verbose_name="email address"),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(default=django.utils.timezone.now, verbose_name="date joined"),
                ),
                (
                    "stripe_customer_id",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("entitlements", models.JSONField(blank=True, default=list, null=True)),
                ("awaiting_email_verification", models.BooleanField(default=True)),
                ("require_change_password", models.BooleanField(default=False)),
                (
                    "role",
                    models.CharField(
                        choices=[
                            ("DEV", "Developer"),
                            ("STAFF", "Staff"),
                            ("USER", "User"),
                            ("TESTER", "Tester"),
                        ],
                        default="USER",
                        max_length=10,
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "abstract": False,
            },
            managers=[
                ("objects", core.models.CustomUserManager()),
            ],
        ),
        migrations.CreateModel(
            name="Error",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("error", models.CharField(max_length=250, null=True)),
                ("error_code", models.CharField(max_length=100, null=True)),
                ("error_colour", models.CharField(default="danger", max_length=25)),
                ("date", models.DateTimeField(auto_now=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="LoginLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "service",
                    models.CharField(
                        choices=[("manual", "Manual"), ("magic_link", "Magic Link")],
                        default="manual",
                        max_length=14,
                    ),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("message", models.CharField(max_length=100)),
                (
                    "action",
                    models.CharField(
                        choices=[
                            ("normal", "Normal"),
                            ("modal", "Modal"),
                            ("redirect", "Redirect"),
                        ],
                        default="normal",
                        max_length=10,
                    ),
                ),
                (
                    "action_value",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("extra_type", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "extra_value",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("date", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="user_notifications",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                (
                    "stripe_customer_id",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("entitlements", models.JSONField(blank=True, default=list, null=True)),
                (
                    "leader",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="teams_leader_of",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "members",
                    models.ManyToManyField(related_name="teams_joined", to=settings.AUTH_USER_MODEL),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AuditLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("action", models.CharField(max_length=300)),
                ("date", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.organization",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="APIAuthToken",
            fields=[
                (
                    "expires",
                    models.DateTimeField(
                        blank=True,
                        help_text="When the item will expire",
                        null=True,
                        verbose_name="Expires",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "hashed_key",
                    models.CharField(max_length=128, unique=True, verbose_name="Key"),
                ),
                ("name", models.CharField(max_length=64, verbose_name="Key Name")),
                (
                    "description",
                    models.TextField(blank=True, null=True, verbose_name="Description"),
                ),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created"),
                ),
                (
                    "last_used",
                    models.DateTimeField(blank=True, null=True, verbose_name="Last Used"),
                ),
                (
                    "scopes",
                    models.JSONField(
                        default=list,
                        help_text="List of permitted scopes",
                        verbose_name="Scopes",
                    ),
                ),
                (
                    "administrator_service_type",
                    models.CharField(
                        blank=True,
                        max_length=64,
                        null=True,
                        verbose_name="Administrator Service Type",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.organization",
                    ),
                ),
            ],
            options={
                "verbose_name": "API Key",
                "verbose_name_plural": "API Keys",
            },
        ),
        migrations.AddField(
            model_name="user",
            name="logged_in_as_team",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.organization",
            ),
        ),
        migrations.CreateModel(
            name="PasswordSecret",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "expires",
                    models.DateTimeField(
                        blank=True,
                        help_text="When the item will expire",
                        null=True,
                        verbose_name="Expires",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("secret", models.TextField(max_length=300)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="password_secrets",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="TeamInvitation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "expires",
                    models.DateTimeField(
                        blank=True,
                        help_text="When the item will expire",
                        null=True,
                        verbose_name="Expires",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("code", models.CharField(max_length=10)),
                (
                    "invited_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "team",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="team_invitations",
                        to="core.organization",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="team_invitations",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Team Invitation",
                "verbose_name_plural": "Team Invitations",
            },
        ),
        migrations.CreateModel(
            name="TracebackError",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("error", models.CharField(max_length=5000, null=True)),
                ("date", models.DateTimeField(auto_now=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserSettings",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("dark_mode", models.BooleanField(default=True)),
                (
                    "currency",
                    models.CharField(
                        choices=[
                            ("GBP", "British Pound Sterling"),
                            ("EUR", "Euro"),
                            ("USD", "United States Dollar"),
                            ("JPY", "Japanese Yen"),
                            ("INR", "Indian Rupee"),
                            ("AUD", "Australian Dollar"),
                            ("CAD", "Canadian Dollar"),
                        ],
                        default="GBP",
                        max_length=3,
                    ),
                ),
                (
                    "profile_picture",
                    models.ImageField(
                        blank=True,
                        null=True,
                        storage=core.models._public_storage,
                        upload_to="profile_pictures/",
                    ),
                ),
                ("disabled_features", models.JSONField(default=list)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="user_profile",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User Settings",
                "verbose_name_plural": "User Settings",
            },
        ),
        migrations.CreateModel(
            name="VerificationCodes",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "expires",
                    models.DateTimeField(
                        blank=True,
                        help_text="When the item will expire",
                        null=True,
                        verbose_name="Expires",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                (
                    "token",
                    models.TextField(default=core.models.RandomCode, editable=False),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "service",
                    models.CharField(
                        choices=[
                            ("create_account", "Create Account"),
                            ("reset_password", "Reset Password"),
                        ],
                        max_length=14,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Verification Code",
                "verbose_name_plural": "Verification Codes",
            },
        ),
        migrations.CreateModel(
            name="EmailSendStatus",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("updated_status_at", models.DateTimeField(auto_now_add=True)),
                ("recipient", models.TextField()),
                (
                    "aws_message_id",
                    models.CharField(blank=True, editable=False, max_length=100, null=True),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("send", "Send"),
                            ("reject", "Reject"),
                            ("bounce", "Bounce"),
                            ("complaint", "Complaint"),
                            ("delivery", "Delivery"),
                            ("open", "Open"),
                            ("click", "Click"),
                            ("rendering_failure", "Rendering_Failure"),
                            ("delivery_delay", "Delivery_Delay"),
                            ("subscription", "Subscription"),
                            ("failed_to_send", "Failed_To_Send"),
                            ("pending", "Pending"),
                        ],
                        max_length=20,
                    ),
                ),
                (
                    "sent_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="emails_sent",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.organization",
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.CheckConstraint(
                        condition=models.Q(
                            models.Q(("organization__isnull", False), ("user__isnull", True)),
                            models.Q(("organization__isnull", True), ("user__isnull", False)),
                            _connector="OR",
                        ),
                        name="core_emailsendstatus_check_user_or_organization",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="TeamMemberPermission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "scopes",
                    models.JSONField(
                        default=list,
                        help_text="List of permitted scopes",
                        verbose_name="Scopes",
                    ),
                ),
                (
                    "team",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="permissions",
                        to="core.organization",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="team_permissions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("team", "user")},
            },
        ),
    ]



from typing import Any

from django import template

register = template.Library()


@register.simple_tag
def list_item_prefix_distinct(items: list, index: int = 0, separator=":"):
    return set(i.split(separator)[index] for i in items)


def lookup_separator_perms(items: list, lookup_value: Any):
    values = [i.split(":")[-1] for i in items if i.split(":")[0] == lookup_value]
    return values


def at_index(items: list, index: int = 0):
    return items[index]


@register.simple_tag
def common_items(*lists: list[list]) -> list:
    result = set(lists[0])
    for lst in lists[1:]:
        result.intersection_update(lst)
    return list(result)


@register.simple_tag
def common_items_count(*lists: list[list]) -> int:
    return len(common_items(*lists))


def common_children_filter(list1, list2):
    return common_items(list1, list2)


@register.filter
def get_first_n_items(value, n):
    """
    Returns the first n items of a list.
    """
    try:
        return value[:n]
    except TypeError:
        return []


register.filter("list_item_prefix_distinct", list_item_prefix_distinct)
register.filter("lookup_separator_perms", lookup_separator_perms)
register.filter("at_index", at_index)
register.filter("common_children_filter", common_children_filter)



from django.core.management.base import BaseCommand
from django.db.models.functions import Length
from django.utils.termcolors import colorize
from core.models import FeatureFlags


class Command(BaseCommand):
    help = "Manage the feature flag statuses"

    def add_arguments(self, parser):
        parser.add_argument("action", type=str, help="enable, disable or list")
        parser.add_argument("flag", type=str, nargs="?", help="feature flag name")
        # parser.add_argument("-f", type=str, dest="flag", help="feature flag name")

    #
    def handle(self, *args, **kwargs):
        if kwargs["action"] == "list":
            flags = FeatureFlags.objects.annotate(name_len=Length("name"), description_len=Length("description"))
            width = flags.order_by("-name_len").first().name_len + 4
            description_width = flags.order_by("-description_len").first().description_len + 4

            header = "{:<{width}} {:<10} {:<{description_width}} {:<20}".format(
                "Name", "Enabled", "Description", "Last updated", width=width, description_width=description_width
            )
            self.stdout.write("Feature flags:")
            self.stdout.write(header)

            for flag in FeatureFlags.objects.all():
                value = "✔" if flag.value else "❌"

                formatted_date = flag.updated_at.strftime("%Y-%m-%d %H:%M:%S")
                row = "{:<{width}} {:<10} {:<{description_width}} {:<20}".format(
                    flag.name, value, flag.description or "No description", formatted_date, width=width, description_width=description_width
                )
                self.stdout.write(row)
            return

        if not kwargs["flag"]:
            self.stdout.write(
                colorize("Please provide a feature flag name with `feature_flags enable|disable <name>`", fg="red", opts=("bold",))
            )
            return

        try:
            flag = FeatureFlags.objects.get(name=kwargs["flag"])

            if kwargs["action"] == "enable":
                flag.enable()
                self.stdout.write(f"[👍] Feature flag {kwargs['flag']} has been enabled")
            elif kwargs["action"] == "disable":
                flag.disable()
                self.stdout.write(f"[👍] Feature flag {kwargs['flag']} has been disabled")
        except FeatureFlags.DoesNotExist:
            self.stdout.write(colorize("Feature flag  with the name of `{kwargs['flag']}` does not exist", fg="red", opts=("bold",)))
            return



import importlib

from django.apps import AppConfig


class CoreConfig(AppConfig):
    name = "core"
    verbose_name = "Strelix Core"

    DEFAULT_AUTHENTICATION_CLASSES = "core.api.public.authentication.CustomBearerAuthentication"

    def ready(self):
        from core.api.public.models import APIAuthToken
        import core.signals

        importlib.import_module("core.service.modals.modals")
        pass



from __future__ import annotations

import json
import os
import sys
from logging import exception

import boto3
import environ
from django_ratelimit.core import get_usage
from mypy_boto3_sesv2.client import SESV2Client
from mypy_boto3_sesv2.type_defs import (
    SendEmailResponseTypeDef,
    BulkEmailEntryTypeDef,
    SendBulkEmailResponseTypeDef,
)

from core.types.emails import (
    SingleEmailInput,
    BulkTemplatedEmailInput,
    SingleTemplatedEmailContent,
    SingleEmailSendServiceResponse,
    BulkEmailSendServiceResponse,
    BulkEmailEmailItem,
)

# NEEDS REFACTOR

env = environ.Env(DEBUG=(bool, False))
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
environ.Env.read_env(os.path.join(BASE_DIR, ".env"))
env = environ.Env()
environ.Env.read_env()


def get_var(key, default=None, required=False):
    value = os.environ.get(key, default=default)

    if required and not value:
        raise ValueError(f"{key} is required")
    if not default and not value:  # So methods like .lower() don't error
        value = ""
    return value


def increment_rate_limit(request, group):
    """
    Alias of is_ratelimited that just increments the rate limit for the given group.

    Returns the new usage count.
    """
    usage = get_usage(request, group, increment=True)
    return usage.get("count", 0)


EMAIL_CLIENT: SESV2Client = boto3.client(
    "sesv2",
    region_name="eu-west-2",
    # aws_access_key_id=get_var("AWS_SES_ACCESS_KEY_ID"),
    # aws_secret_access_key=get_var("AWS_SES_SECRET_ACCESS_KEY"),
)

# AWS_SES_ACCESS_KEY_ID = get_var("AWS_SES_ACCESS_KEY_ID")
# AWS_SES_SECRET_ACCESS_KEY = get_var("AWS_SES_SECRET_ACCESS_KEY")
AWS_SES_FROM_ADDRESS = get_var("AWS_SES_FROM_ADDRESS")
ARE_AWS_EMAILS_ENABLED = (True if get_var("AWS_SES_ENABLED", "").lower() == "true" else False) and AWS_SES_FROM_ADDRESS

# SENDGRID_TEMPLATE = get_var("SENDGRID_TEMPLATE")
EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"
# EMAIL_BACKEND = "sendgrid_backend.SendgridBackend"
# EMAIL_BACKEND = "sendgrid_backend.SendgridBackend"
# EMAIL_HOST = "smtp.sendgrid.net"
# EMAIL_HOST_USER = "apikey"
# EMAIL_FROM_ADDRESS = get_var("SENDGRID_FROM_ADDRESS")
# EMAIL_HOST_PASSWORD = get_var("SENDGRID_API_KEY")
# EMAIL_PORT = 587
# EMAIL_USE_TLS = True
# EMAIL_SERVER_ENABLED = True if EMAIL_HOST_PASSWORD else False

EMAIL_SERVICE = "SES" if ARE_AWS_EMAILS_ENABLED else None
ARE_EMAILS_ENABLED = True if EMAIL_SERVICE else False

if "test" in sys.argv[1:]:
    ARE_EMAILS_ENABLED = False


def send_email(
    destination: str | list[str],
    subject: str | None,
    content: str | SingleTemplatedEmailContent,
    ConfigurationSetName: str | None = None,
    from_address: str | None = None,
    from_address_name_prefix: str | None = None,
) -> SingleEmailSendServiceResponse:
    """
    Args:
    destination (email addr or list of email addr): The email address or list of email addresses to send the
    email to.
    subject (str): The subject of the email.
    message (str): The content of the email.
    """

    data = SingleEmailInput(
        destination=destination,
        subject=subject,
        content=content,
        ConfigurationSetName=ConfigurationSetName,
        from_address=from_address,
        from_address_name_prefix=from_address_name_prefix,
    )

    if get_var("DEBUG", "").lower() == "true":
        if not "test" in sys.argv[1:]:
            print(data)
        return SingleEmailSendServiceResponse(
            True,
            response=SendEmailResponseTypeDef(
                MessageId="",
                ResponseMetadata={
                    "RequestId": "",
                    "HTTPStatusCode": 200,
                    "HTTPHeaders": {},
                    "RetryAttempts": 0,
                    "HostId": "",
                },
            ),
        )

    if EMAIL_SERVICE == "SES":
        if not isinstance(data.destination, list):
            data.destination = [data.destination]

        response: SendEmailResponseTypeDef | None = None

        try:
            if isinstance(data.content, dict):
                data_str = (
                    data.content.get("template_data")
                    if isinstance(data.content.get("template_data"), str)
                    else json.dumps(data.content.get("template_data"))
                )

                from_email_address: str = str(data.from_address_name_prefix) if data.from_address_name_prefix else ""
                from_email_address += str(data.from_address or AWS_SES_FROM_ADDRESS)

                response = EMAIL_CLIENT.send_email(
                    FromEmailAddress=from_email_address,
                    Destination={"ToAddresses": data.destination},
                    Content={"Template": {"TemplateName": data.content.get("template_name"), "TemplateData": data_str}},  # type: ignore
                    ConfigurationSetName=data.ConfigurationSetName or "",
                )
            else:
                from_email_address = str(data.from_address_name_prefix) if data.from_address_name_prefix else ""
                from_email_address += str(data.from_address or AWS_SES_FROM_ADDRESS)

                response = EMAIL_CLIENT.send_email(
                    FromEmailAddress=from_email_address,
                    Destination={"ToAddresses": data.destination},
                    Content={
                        "Simple": {"Subject": {"Data": data.subject if data.subject else ""}, "Body": {"Text": {"Data": data.content}}}
                    },
                    ConfigurationSetName=data.ConfigurationSetName or "",
                )
            return SingleEmailSendServiceResponse(True, response=response)
        except EMAIL_CLIENT.exceptions.MessageRejected:
            return SingleEmailSendServiceResponse(error_message="Email rejected", response=response)

        except EMAIL_CLIENT.exceptions.AccountSuspendedException:
            return SingleEmailSendServiceResponse(error_message="Email account suspended", response=response)

        except EMAIL_CLIENT.exceptions.SendingPausedException:
            return SingleEmailSendServiceResponse(error_message="Email sending paused", response=response)

        except Exception as error:
            exception(f"Unexpected error occurred: {error}")
            return SingleEmailSendServiceResponse(error_message="Email service error", response=response)
    return SingleEmailSendServiceResponse(error_message="No email service configured")


def send_bulk_email(
    email_list: list[BulkEmailEmailItem],
    ConfigurationSetName: str | None = None,
    from_address: str | None = None,
) -> BulkEmailSendServiceResponse:

    entries: list[BulkEmailEntryTypeDef] = [
        {
            "Destination": {
                "ToAddresses": [entry.destination] if not isinstance(entry.destination, list) else entry.destination,
                "CcAddresses": entry.cc,
                "BccAddresses": entry.bcc,
            }
        }
        for entry in email_list
    ]

    try:
        response: SendBulkEmailResponseTypeDef = EMAIL_CLIENT.send_bulk_email(
            FromEmailAddress=from_address or AWS_SES_FROM_ADDRESS,
            BulkEmailEntries=entries,
            ConfigurationSetName=ConfigurationSetName or "",
            DefaultContent={},
        )

        return BulkEmailSendServiceResponse(True, response=response)
    except EMAIL_CLIENT.exceptions.MessageRejected:
        return BulkEmailSendServiceResponse(error_message="Email rejected", response=locals().get("response", None))

    except EMAIL_CLIENT.exceptions.AccountSuspendedException:
        return BulkEmailSendServiceResponse(error_message="Email account suspended", response=locals().get("response", None))

    except EMAIL_CLIENT.exceptions.SendingPausedException:
        return BulkEmailSendServiceResponse(error_message="Email sending paused", response=locals().get("response", None))

    except Exception as error:
        exception(f"Unexpected error occurred: {error}")
        return BulkEmailSendServiceResponse(error_message="Email service error", response=locals().get("response", None))


def send_templated_bulk_email(
    email_list: list[BulkEmailEmailItem],
    template_name: str,
    default_template_data: dict | str,
    ConfigurationSetName: str | None = None,
    from_address: str | None = None,
    from_address_name_prefix: str | None = None,
) -> BulkEmailSendServiceResponse:

    data = BulkTemplatedEmailInput(
        email_list=email_list,
        template_name=template_name,
        default_template_data=default_template_data,
        ConfigurationSetName=ConfigurationSetName,
        from_address=from_address,
        from_address_name_prefix=from_address_name_prefix,
    )

    entries: list[BulkEmailEntryTypeDef] = []

    for entry in data.email_list:
        destination: list[str] = [entry.destination] if not isinstance(entry.destination, list) else entry.destination

        data_str: str = entry.template_data if isinstance(entry.template_data, str) else json.dumps(entry.template_data)

        entries.append(
            {
                "Destination": {"ToAddresses": destination, "CcAddresses": entry.cc, "BccAddresses": entry.bcc},
                "ReplacementEmailContent": {"ReplacementTemplate": {"ReplacementTemplateData": data_str}},
            }
        )

    try:
        data_str = data.default_template_data if isinstance(data.default_template_data, str) else json.dumps(data.default_template_data)
        from_email_address: str = str(data.from_address_name_prefix) if data.from_address_name_prefix else ""
        from_email_address += str(data.from_address or AWS_SES_FROM_ADDRESS)

        response: SendBulkEmailResponseTypeDef = EMAIL_CLIENT.send_bulk_email(
            FromEmailAddress=from_email_address,
            BulkEmailEntries=entries,
            ConfigurationSetName=data.ConfigurationSetName or "",
            DefaultContent={"Template": {"TemplateName": data.template_name, "TemplateData": data_str}},
        )
        return BulkEmailSendServiceResponse(True, response=response)
    except EMAIL_CLIENT.exceptions.MessageRejected:
        return BulkEmailSendServiceResponse(error_message="Email rejected", response=locals().get("response", None))

    except EMAIL_CLIENT.exceptions.AccountSuspendedException:
        return BulkEmailSendServiceResponse(error_message="Email account suspended", response=locals().get("response", None))

    except EMAIL_CLIENT.exceptions.SendingPausedException:
        return BulkEmailSendServiceResponse(error_message="Email sending paused", response=locals().get("response", None))

    except Exception as error:
        exception(f"Unexpected error occurred: {error}")
        return BulkEmailSendServiceResponse(error_message="Email service error", response=locals().get("response", None))


print(sys.argv)
if not any(arg in sys.argv[1:] for arg in ["test", "migrate", "makemigrations"]) and sys.argv != ["makemigrations.py"]:
    if not get_var("SITE_URL"):
        raise ValueError("SITE_URL is required")

    if not get_var("SITE_NAME"):
        raise ValueError("SITE_NAME is required")



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal


@dataclass
class QuotaItem:
    slug: str
    name: str
    description: str
    default_value: int
    adjustable: bool
    period: Literal[
        "forever",
        "per_month",
        "per_minute",
        "per_hour",
        "per_day",
        "per_client",
        "per_invoice",
        "per_team",
        "per_quota",
        "per_bulk_send",
        "per_email",
    ]


@dataclass
class QuotaGroup:
    name: str
    items: list[QuotaItem]


default_quota_limits: list[QuotaGroup] = [
    QuotaGroup(
        "invoices",
        [
            QuotaItem(
                slug="count",
                name="Creations per month",
                description="Amount of invoices created per month",
                default_value=100,
                period="per_month",
                adjustable=True,
            ),
            QuotaItem(
                slug="schedules",
                name="Schedules per month",
                description="Amount of invoice scheduled sends allowed per month",
                default_value=100,
                period="per_month",
                adjustable=True,
            ),
            QuotaItem(
                slug="access_codes",
                name="Created access codes",
                description="Amount of invoice access codes allowed per invoice",
                default_value=3,
                period="per_invoice",
                adjustable=True,
            ),
        ],
    ),
    QuotaGroup(
        "receipts",
        [
            QuotaItem(
                slug="count",
                name="Created receipts",
                description="Amount of receipts stored per month",
                default_value=100,
                period="per_month",
                adjustable=True,
            )
        ],
    ),
    QuotaGroup(
        "clients",
        [
            QuotaItem(
                slug="count",
                name="Created clients",
                description="Amount of clients stored in total",
                default_value=40,
                period="forever",
                adjustable=True,
            )
        ],
    ),
    QuotaGroup(
        "teams",
        [
            QuotaItem(
                slug="count",
                name="Created teams",
                description="Amount of teams created in total",
                default_value=3,
                period="forever",
                adjustable=True,
            ),
            QuotaItem(
                slug="joined",
                name="Joined teams",
                description="Amount of teams that you have joined in total",
                default_value=5,
                period="forever",
                adjustable=True,
            ),
            QuotaItem(
                slug="user_count",
                name="Users per team",
                description="Amount of users per team",
                default_value=10,
                period="per_team",
                adjustable=True,
            ),
        ],
    ),
    QuotaGroup(
        "quota_increase",
        [
            QuotaItem(
                slug="request",
                name="Quota Increase Request",
                description="Amount of increase requests allowed per quota",
                default_value=1,
                period="per_quota",
                adjustable=False,
            ),
            QuotaItem(
                slug="requests_per_month_per_quota",
                name="Quota Increase Requests per month",
                description="Amount of increase requests allowed per month per quota",
                period="per_quota",
                default_value=1,
                adjustable=False,
            ),
        ],
    ),
    QuotaGroup(
        "emails",
        [
            QuotaItem(
                slug="single-count",
                name="Single Email Sends",
                description="Amount of single email sends allowed per month",
                period="per_month",
                default_value=10,
                adjustable=True,
            ),
            QuotaItem(
                slug="bulk-count",
                name="Bulk Email Sends",
                description="Amount of 'Bulk Emails' allowed to be sent per month",
                period="per_month",
                default_value=1,
                adjustable=True,
            ),
            QuotaItem(
                slug="bulk-max_sends",
                name="Bulk Email Maximum Emails",
                description="Maximum amount of emails allowed to be sent per 'Bulk' request",
                period="per_bulk_send",
                default_value=10,
                adjustable=True,
            ),
            QuotaItem(
                slug="email_character_count",
                name="Maximum Character Count",
                description="Maximum amount of characters allowed in an email",
                period="per_email",
                default_value=1000,
                adjustable=True,
            ),
            QuotaItem(
                slug="complaints",
                name="Complaints allowed",
                description="Maximum amount of complaints allowed before your account will be blocked from sending emails",
                period="forever",
                default_value=2,
                adjustable=True,
            ),
        ],
    ),
]



import logging
from datetime import datetime

import stripe
from django.db.models.signals import post_save
from django.dispatch import receiver

from core.models import User
from billing.models import BillingUsage

logger = logging.getLogger(__name__)


@receiver(post_save, sender=BillingUsage)
def usage_occurred(sender, instance: BillingUsage, created, **kwargs):
    if not created or instance.processed:
        return

    if instance.event_type != "usage":
        return  # may add storage at a later point

    if not instance.owner:
        print("CANNOT HANDLE ORGS AT THE MOMENT!")
        return  # todo: cannot handle organisations at the moment

    stripe_customer_id = instance.owner.stripe_customer_id

    if not stripe_customer_id:
        print(f"No stripe customer id for actor #{'usr_' if isinstance(instance.owner, User) else 'org_'}{instance.owner.id}")
        return  # todo

    meter_event = stripe.billing.MeterEvent.create(
        event_name=instance.event_name, payload={"value": str(instance.quantity), f"stripe_customer_id": stripe_customer_id}
    )

    if meter_event.created:
        instance.stripe_unique_usage_identifier = meter_event.identifier
        instance.set_processed(datetime.fromtimestamp(meter_event.created))

    return



from django.core.management.base import BaseCommand
from django.core.cache import cache
import stripe


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "action",
            type=str,
            help="help, create_entitlements",
        )

    def handle(self, *args, **kwargs):
        action = kwargs.get("action")

        match action:
            case "create_entitlements":
                for entitlement in [
                    "Receipts",
                    "File Storage",
                    "Organizations",
                    "Invoice Reminders",
                    "API Access",
                    "Emails",
                    "Advanced Onboarding",
                    "Basic Onboarding",
                    "Invoice Schedules",
                    "Invoices",
                    "Customers",
                ]:
                    try:
                        stripe.entitlements.Feature.create(name=entitlement, lookup_key=entitlement.lower().replace(" ", "-"))
                        print(f"Created entitlement: {entitlement}")
                    except stripe.error.InvalidRequestError:
                        print(f"Entitlement already exists: {entitlement}")
            case (_, "help"):
                print(
                    """
                    Available actions:
                    - create_entitlements: This will create all entitlements that you don't have
                """
                )



import traceback

from django.contrib import messages
from django.http import HttpRequest
from django.shortcuts import redirect
from django_ratelimit.exceptions import Ratelimited

from core.models import TracebackError, AuditLog


def universal(request: HttpRequest, exception=None):
    messages.error(
        request,
        "Sorry, something went wrong on our end! We've contacted our team, please email us if this issue continues.",
    )
    traceback.print_exc()
    exec_error = traceback.format_exc()
    print(f"WAS A TRACEBACK ERROR: EXCEPTION: {exception}")

    if len(exec_error) > 4999:
        return

    if request.user.is_authenticated:
        messages.error(request, "Sorry, something went wrong!")
        TracebackError(user=request.user, error=exec_error).save()
    else:
        TracebackError(error=exec_error).save()

    return redirect("dashboard")


def e_403(request: HttpRequest, exception=None):
    if isinstance(exception, Ratelimited):
        messages.error(
            request,
            "Woah, slow down there. You've been temporarily blocked from this page due to extreme requests.",
        )
        user_ip = request.META.get("REMOTE_ADDR")
        user_id = f"User #{request.user.id}" if request.user.is_authenticated else "Not logged in"
        action = f"{user_ip} | Ratelimited | {user_id}"
        auditlog = AuditLog(action=action)
        if request.user.is_authenticated:
            auditlog.user = request.user
        auditlog.save()
        return redirect("core:auth:login")
    else:
        messages.error(
            request,
            "Sorry, something went wrong on our end!" "We've contacted our team, please email us if this issue continues.",
        )
        traceback.print_exc()
        exec_error = traceback.format_exc()
        print(f"WAS A TRACEBACK ERROR: EXCEPTION: {exception}")

        if len(exec_error) > 4999:
            return

        if request.user.is_authenticated:
            TracebackError(user=request.user, error=exec_error).save()
        else:
            TracebackError(error=exec_error).save()

        return redirect("dashboard")



from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.db.models import Q
from django.shortcuts import redirect, render
from django.views import View

from core.models import User
from core.utils.feature_flags import get_feature_status


class CreateAccountChooseView(View):
    def get(self, request):
        if request.user.is_authenticated:
            return redirect("dashboard")
        SIGNUPS_ENABLED = get_feature_status("areSignupsEnabled")
        if not SIGNUPS_ENABLED:
            messages.error(request, "New account signups are currently disabled")
            return redirect("core:auth:login")
        return render(
            request,
            "core/auth/create_account_choose.html",
            {
                "github_enabled": getattr(settings, "SOCIAL_AUTH_GITHUB_ENABLED", False),
                "google_enabled": getattr(settings, "SOCIAL_AUTH_GOOGLE_OAUTH2_ENABLED", False),
            },
        )


class CreateAccountManualView(View):
    def get(self, request):
        if request.user.is_authenticated:
            return redirect("dashboard")
        SIGNUPS_ENABLED = get_feature_status("areSignupsEnabled")
        if not SIGNUPS_ENABLED:
            messages.error(request, "New account signups are currently disabled")
            return redirect("core:auth:login")
        return render(request, "core/auth/create_account_manual.html")

    def post(self, request):
        if request.user.is_authenticated:
            return redirect("dashboard")
        SIGNUPS_ENABLED = get_feature_status("areSignupsEnabled")
        if not SIGNUPS_ENABLED:
            messages.error(request, "New account signups are currently disabled")
            return redirect("core:auth:login")

        email = request.POST.get("email")
        password = request.POST.get("password")
        password_confirm = request.POST.get("confirm_password")

        if password != password_confirm:
            messages.error(request, "Passwords don't match")
            return render(
                request,
                "core/auth/create_account_manual.html",
                {"attempted_email": email},
            )

        try:
            validate_email(email)
        except ValidationError:
            messages.error(request, "Invalid email")
            return render(request, "core/auth/create_account_manual.html")

        emails_taken = User.objects.filter(Q(username=email) | Q(email=email)).exists()

        if emails_taken:
            messages.error(request, "Email is already taken")
            return render(request, "core/auth/create_account_manual.html")

        if len(password) < 6:
            messages.error(request, "Password must be at least 6 characters")
            return render(request, "core/auth/create_account_manual.html")

        created_user = User.objects.create_user(email=email, username=email, password=password)
        created_user.awaiting_email_verification = True
        created_user.save()
        # created_user.is_active = False
        # created_user.save()

        user = authenticate(request, username=email, password=password)
        if not user:
            messages.error(request, "Something went wrong")
            return render(request, "core/auth/create_account_manual.html")

        # login(request, user)
        messages.success(
            request,
            "Successfully created account. Please verify your account via the email we are " "sending you now!",
        )
        return redirect("core:auth:login")



from django.db.models.signals import post_save
from django.dispatch import receiver
from billing.models import StripeWebhookEvent
from billing.service.checkout_completed import checkout_completed
from billing.service.subscription_ended import subscription_ended
from billing.service.entitlements import entitlements_updated_via_stripe_webhook


@receiver(post_save, sender=StripeWebhookEvent)
def stripe_webhook_event_created(sender, instance: StripeWebhookEvent, **kwargs):
    match instance.event_type:
        case "checkout.session.completed":
            checkout_completed(instance)
        case "customer.subscription.deleted":
            subscription_ended(instance)
        case "entitlements.active_entitlement_summary.updated":
            entitlements_updated_via_stripe_webhook(instance)
        case _:
            print(f"Unhandled event type: {instance.event_type}")



from django.urls import path
from .webhook_task_queue_handler import webhook_task_queue_handler_view_endpoint

urlpatterns = [
    path(
        "receive/global/",
        webhook_task_queue_handler_view_endpoint,
        name="receive_global",
    )
]

app_name = "webhooks"



# from django.db.models.signals import post_save
# from django.dispatch import receiver
#
# from backend.finance.models import Invoice, Usage
#
#
# @receiver(post_save, sender=Invoice)
# def created_invoice(sender, instance: Invoice, **kwargs):
#     Usage.objects.create(
#         owner=instance.owner,
#         feature="invoices-created",
#         quantity=1,
#         unit="invocations",
#         instance_id=instance.id,
#     )



from django.core.management.base import BaseCommand
from django.core.management import call_command


class Command(BaseCommand):
    help = "Runs URL verification tests."

    def handle(self, *args, **options):
        call_command("test", "backend.tests.urls.verify_urls")



from django.urls import path

# from . import fetch, requests

urlpatterns = [
    # path(
    #     "fetch/<str:group>/",
    #     fetch.fetch_all_quotas,
    #     name="fetch",
    # ),
    # path("submit_request/<slug:slug>/", requests.submit_request, name="submit_request"),
    # path("request/<int:request_id>/approve/", requests.approve_request, name="approve request"),
    # path("request/<int:request_id>/decline/", requests.decline_request, name="decline request"),
]

app_name = "quotas"



from django.http import HttpRequest
from django.shortcuts import render
from login_required import login_not_required


def index(request: HttpRequest):
    return render(request, "pages/landing/index.html")


@login_not_required
def pricing(request: HttpRequest):
    return render(request, "pages/landing/pricing.html")


def dashboard(request: HttpRequest):
    return render(request, "pages/dashboard.html")



import logging

from django import template
from django.urls import NoReverseMatch

from core.models import User, Organization
from core.utils.feature_flags import get_feature_status

from django.conf import settings

register = template.Library()


@register.simple_tag
def has_module(module_str: str):
    return module_str in settings.INSTALLED_APPS


@register.simple_tag
def safe_url(view_name, *args, **kwargs):
    from django.urls import reverse

    try:
        return reverse(view_name, args=args, kwargs=kwargs)
    except NoReverseMatch:
        return ""


@register.simple_tag
def feature_enabled(feature):
    return get_feature_status(feature)


@register.simple_tag
def personal_feature_enabled(user: User, feature: str):
    return user.user_profile.has_feature(feature)


@register.simple_tag
def has_entitlement(actor: User | Organization, entitlement: str) -> bool:
    if not hasattr(settings, "BILLING_ENABLED") or not settings.BILLING_ENABLED:
        return True

    from billing.service.entitlements import has_entitlement as _has_entitlement

    return _has_entitlement(actor, entitlement)



import stripe

from core.models import User, Organization
from billing.models import StripeWebhookEvent, UserSubscription


def subscription_ended(webhook_event: StripeWebhookEvent) -> None:
    event_data: stripe.Subscription = webhook_event.data.object
    stripe_customer = event_data.customer

    # Find the user or organization based on the stripe customer
    actor = (
        User.objects.filter(stripe_customer_id=stripe_customer).first()
        or Organization.objects.filter(stripe_customer_id=stripe_customer).first()
    )

    actor_subscription_plan = None

    if not actor:
        # If no user found, try to fetch the subscription plan using the stripe subscription ID
        plan = UserSubscription.objects.filter(
            stripe_subscription_id=event_data.id, stripe_subscription_id__isnull=False
        ).first()  # type: ignore[misc]

        if plan:
            actor_subscription_plan = plan
            actor = plan.owner
        else:
            print("Error: Could not find user or subscription plan.")
            return

    if not actor_subscription_plan:
        actor_subscriptions = UserSubscription.filter_by_owner(owner=actor).all()
        if not actor_subscriptions:
            return

        # Find a subscription plan with the same Stripe subscription ID
        if plan_with_same_id := actor_subscriptions.filter(stripe_subscription_id=event_data.id).first():
            plan_with_same_id.end_now()



from django.shortcuts import render

from core.types.requests import WebRequest
from core.service.base.breadcrumbs import get_breadcrumbs


def update_breadcrumbs_endpoint(request: WebRequest):
    url = request.GET.get("url")

    breadcrumb_dict: dict = get_breadcrumbs(url=url)
    return render(
        request,
        "core/base/breadcrumbs.html",
        {
            "breadcrumb": breadcrumb_dict.get("breadcrumb"),
            "swapping": True,
            # "swap": True
        },
    )



from core.models import FeatureFlags
from django.core.cache import cache
from django.core.cache.backends.redis import RedisCacheClient

cache: RedisCacheClient = cache


def get_feature_status(feature, should_use_cache=True):
    if should_use_cache:
        key = f"myfinances:feature_flag:{feature}"
        cached_value = cache.get(key)
        if cached_value:
            return cached_value

    value = FeatureFlags.objects.filter(name=feature).first()
    if value:
        if should_use_cache:
            cache.set(key, value.value, timeout=300)
        return value.value
    else:
        return False


def set_cache(key, value, timeout=300):
    cache.set(key, value, timeout=timeout)


def get_cache(key):
    return cache.get(key)



from django.core.management.base import BaseCommand
from django.core.cache import cache


class Command(BaseCommand):
    """
    Deletes the "navbar_items" cache and prints a message to the standard output.
    """

    def handle(self, *args, **kwargs):
        cache.delete("navbar_items")
        self.stdout.write("Cleared cache\n")



from django.urls import path
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions

INFO = openapi.Info(
    title="MyFinances Public API",
    default_version="v0.0.1",
    description="",
    terms_of_service="",
    contact=openapi.Contact(email="support@strelix.org"),
    license=openapi.License(name="AGPL v3"),
)

schema_view = get_schema_view(
    INFO,
    public=True,
    permission_classes=[permissions.AllowAny],
)


def get_swagger_ui():
    return schema_view


def get_swagger_endpoints(debug):
    return (
        [
            path("redoc/", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"),
        ]
        + [
            path("swagger<format>/", schema_view.without_ui(cache_timeout=0), name="schema-json"),
            path("swagger/", schema_view.with_ui("swagger", cache_timeout=0), name="schema-swagger-ui"),
        ]
        if debug
        else []
    )


TEAM_PARAMETER = openapi.Parameter(
    "team_id", openapi.IN_QUERY, description="id of the team you want to do this action under", type=openapi.TYPE_STRING, required=False
)

API_RESPONSE_BASE = lambda resp: {
    "meta": openapi.Schema(
        type=openapi.TYPE_OBJECT,
        properties={
            "success": openapi.Schema(type=openapi.TYPE_BOOLEAN, description="Request served successfully"),
        },
    ),
    "data": openapi.Schema(type=openapi.TYPE_OBJECT, properties=resp),
}



from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from django.conf import settings

SCOPES = {
    "clients:read",
    "clients:write",
    "invoices:read",
    "invoices:write",
    "receipts:read",
    "receipts:write",
    "clients:read",
    "clients:write",
    "emails:read",
    "emails:send",
    "profile:read",
    "profile:write",
    "api_keys:read",
    "api_keys:write",
    "team_permissions:read",
    "team_permissions:write",
    "team:invite",
    "team:kick",
    "account_defaults:write",
    "email_templates:read",
    "email_templates:write",
}

SCOPES_TREE = {
    "clients:read": {"clients:read"},
    "clients:write": {"clients:read", "clients:write"},
    "invoices:read": {"invoices:read"},
    "invoices:write": {"invoices:read", "invoices:write"},
    "profile:read": {"profile:read"},
    "profile:write": {"profile:read", "profile:write"},
    "api_keys:read": {"api_keys:read"},
    "api_keys:write": {"api_keys:read", "api_keys:write"},
    "team_permissions:read": {"team_permissions:read"},
    "team_permissions:write": {"team_permissions:read", "team_permissions:write"},
    "team:invite": {"team:invite"},
    "team:kick": {"team:kick", "team:invite"},
    "email_templates:read": {"email_templates:read"},
    "email_templates:write": {"email_templates:read", "email_templates:write"},
    "account_defaults:write": {"account_defaults:write"},
}

SCOPE_DESCRIPTIONS = {
    "clients": {"description": "Access customer details", "options": {"read": "Read only", "write": "Read and write"}},
    "invoices": {"description": "Access invoices", "options": {"read": "Read only", "write": "Read and write"}},
    "profile": {"description": "Access profile details", "options": {"read": "Read only", "write": "Read and write"}},
    "api_keys": {"description": "Access API keys", "options": {"read": "Read only", "write": "Read and write"}},
    "team_permissions": {"description": "Access team permissions", "options": {"read": "Read only", "write": "Read and write"}},
    "team": {"description": "Invite team members", "options": {"invite": "Invite members"}},
    "email_templates": {"description": "Access email templates", "options": {"read": "Read only", "write": "Read and write"}},
    "account_defaults": {"description": "Modify account defaults", "options": {"write": "Read and write"}},
}

if hasattr(settings, "BILLING_ENABLED") and settings.BILLING_ENABLED:
    SCOPES.add("billing:manage")
    SCOPES_TREE["billing:manage"] = {"billing:manage"}
    SCOPE_DESCRIPTIONS["billing"] = {"description": "Access billing details + stripe", "options": {"manage": "Manage billing"}}


class IsSuperuser(BasePermission):
    def has_permission(self, request: Request, view: object) -> bool:
        return bool(request.user and request.user.is_superuser)



from django.core.management.base import BaseCommand
from django.template.utils import get_app_template_dirs


class Command(BaseCommand):
    help = "List core template files"

    def handle(self, *args, **options):
        app_template_dirs = get_app_template_dirs("templates")
        app_template_dirs = map(lambda a: "TMPL:" + str(a), app_template_dirs)
        self.stdout.write("\n".join(app_template_dirs))



import stripe
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.decorators import api_view, authentication_classes, permission_classes
from billing.billing_settings import STRIPE_WEBHOOK_ENDPOINT_SECRET
from billing.models import StripeWebhookEvent
from django.views.decorators.csrf import csrf_exempt


@api_view(["POST"])
@authentication_classes([])  # No auth required for webhooks
@permission_classes([AllowAny])
@csrf_exempt
def stripe_listener_webhook_endpoint(request: Request):
    payload = request.body
    sig_header = request.META.get("HTTP_STRIPE_SIGNATURE", "")

    try:
        event = stripe.Webhook.construct_event(payload, sig_header, STRIPE_WEBHOOK_ENDPOINT_SECRET)
        print(f"Webhook received: {event['type']}")
    except ValueError as error:
        print(f"Invalid payload: {error}")
        return Response(status=400)
    except stripe.error.SignatureVerificationError as error:
        print(f"Invalid signature: {error}")
        return Response(status=400)

    # Store event in database
    StripeWebhookEvent.objects.create(event_id=event.id, event_type=event["type"], data=event["data"], raw_event=event)

    # Call specific event handler (signal)
    return Response(status=200)



import datetime

from django.utils import timezone


def get_months_text() -> list[str]:
    return [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]


def timezone_now() -> datetime.datetime:
    return timezone.now()



from __future__ import annotations

from django.conf.urls import include
from django.urls import path, re_path
from rest_framework.authentication import TokenAuthentication

from .endpoints.system_health import system_health_endpoint

INTERNAL_URLS = [path("health/", system_health_endpoint, name="public-system-health")]

urlpatterns = [
    path("internal/", include(INTERNAL_URLS)),
    # path("clients/", include("core.api.public.endpoints.clients.urls")),
    # path("invoices/", include("core.api.public.endpoints.Invoices.urls")),
    path("webhooks/", include("core.api.public.endpoints.webhooks.urls")),
]

app_name = "public"



import uuid
from django.core.management.base import BaseCommand
from core.api.public import APIAuthToken


class Command(BaseCommand):
    """
    Generates an API key for the AWS EventBridge API.
    """

    def handle(self, *args, **kwargs):
        token = APIAuthToken(service=APIAuthToken.AdministratorServiceTypes.AWS_API_DESTINATION, name=str(uuid.uuid4()))
        raw_key: str = token.generate_key()
        token.save()

        self.stdout.write(
            f"""
        NOTE: Keep this key secret. It is used to authenticate your API requests with the AWS EventBridge API.

        Your API Key: {raw_key}

        To use this API Key for development you can use:

        pulumi config set api_destination-api_key {raw_key}
        pulumi up

        If you would like to use it for production use:
        pulumi stack select production
        pulumi config set api_destination-api_key {raw_key}
        pulumi up
        """
        )



import stripe
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse, resolve, NoReverseMatch

from core.decorators import web_require_scopes
from core.types.requests import WebRequest

from billing.service.stripe_customer import get_or_create_customer_id


@web_require_scopes("billing:manage", api=True, htmx=True)
def customer_client_portal_endpoint(request: WebRequest):
    if NEXT := request.GET.get("back"):
        try:
            resolve(NEXT)
        except NoReverseMatch:
            NEXT = None

    customer_id = get_or_create_customer_id(request.actor)

    stripe_resp = stripe.billing_portal.Session.create(
        customer=customer_id, return_url=request.build_absolute_uri(NEXT or reverse("dashboard"))
    )

    if request.htmx:
        response = HttpResponse(status=200)
        response["HX-Redirect"] = stripe_resp.url
        return response

    return HttpResponseRedirect(stripe_resp.url)



from django.contrib import messages
from django.shortcuts import redirect, render
from django.views.decorators.http import require_http_methods

from core.service.settings.update import update_profile_picture, UpdateProfilePictureServiceResponse
from core.service.settings.view import get_user_profile
from core.types.requests import WebRequest


@require_http_methods(["POST"])
def change_profile_picture_endpoint(request: WebRequest):
    if not request.htmx:
        messages.error(request, "Invalid request")
        return redirect("core:settings:dashboard with page", page="profile")

    user_profile = get_user_profile(request)

    update_response: UpdateProfilePictureServiceResponse = update_profile_picture(request.FILES.get("profile_picture_image"), user_profile)

    if update_response.success:
        messages.success(request, update_response.response)
    else:
        messages.error(request, update_response.error)

    return render(
        request,
        "pages/settings/settings/_post_profile_pic.html",
        {"users_profile_picture": user_profile.profile_picture_url},
    )



from django.db import models
from django.contrib.auth.hashers import check_password, make_password
import binascii
import os
from django.utils import timezone

from core.models import OwnerBase, ExpiresBase


class APIAuthToken(OwnerBase, ExpiresBase):
    id = models.AutoField(primary_key=True)

    hashed_key = models.CharField("Key", max_length=128, unique=True)

    name = models.CharField("Key Name", max_length=64)
    description = models.TextField("Description", blank=True, null=True)
    created = models.DateTimeField("Created", auto_now_add=True)
    last_used = models.DateTimeField("Last Used", null=True, blank=True)
    # expires = models.DateTimeField("Expires", null=True, blank=True, help_text="Leave blank for no expiry")
    # expired = models.BooleanField("Expired", default=False, help_text="If the key has expired")
    # active = models.BooleanField("Active", default=True, help_text="If the key is active")
    scopes = models.JSONField("Scopes", default=list, help_text="List of permitted scopes")

    class AdministratorServiceTypes(models.TextChoices):
        AWS_WEBHOOK_CALLBACK = "aws_webhook_callback", "AWS Webhook Callback"
        AWS_API_DESTINATION = "aws_api_destination", "AWS API Destination"

    administrator_service_type = models.CharField("Administrator Service Type", max_length=64, blank=True, null=True)

    class Meta:
        verbose_name = "API Key"
        verbose_name_plural = "API Keys"
        app_label = "core"

    def __str__(self):
        return self.name

    def update_last_used(self):
        self.last_used = timezone.now()
        self.save()
        return True

    # def save(self, *args, **kwargs):
    #     return super().save(*args, **kwargs)

    def generate_key(self) -> str:
        """
        :returns: raw_key
        """

        raw = binascii.hexlify(os.urandom(20)).decode()
        self.hashed_key = self.hash_raw_key(raw)

        return raw

    @classmethod
    def hash_raw_key(cls, raw_key: str):
        return make_password(raw_key, salt="api_tokens", hasher="default")

    def verify(self, key) -> bool:
        return check_password(key, self.hashed_key)

    def deactivate(self):
        self.active = False
        self.save()
        return self

    def has_scope(self, scope):
        return scope in self.scopes



# Generated by Django 5.1.4 on 2024-12-21 22:26

import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="BillingUsage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("event_name", models.CharField(max_length=100)),
                (
                    "event_type",
                    models.CharField(
                        choices=[("usage", "Metered Usage")],
                        default="usage",
                        max_length=20,
                    ),
                ),
                ("quantity", models.PositiveSmallIntegerField(default=1)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("processed_at", models.DateTimeField(blank=True, null=True)),
                ("processed", models.BooleanField(default=False)),
                (
                    "stripe_unique_usage_identifier",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="PlanFeature",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("slug", models.CharField(max_length=100)),
                (
                    "stripe_price_id",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "description",
                    models.TextField(blank=True, max_length=500, null=True),
                ),
                ("max_limit_per_month", models.IntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="PlanFeatureGroup",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name="StripeCheckoutSession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4, unique=True)),
                (
                    "stripe_session_id",
                    models.CharField(blank=True, max_length=100, null=True, unique=True),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="StripeWebhookEvent",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("event_id", models.CharField(max_length=100, unique=True)),
                ("event_type", models.CharField(max_length=100)),
                ("data", models.JSONField()),
                ("raw_event", models.JSONField()),
            ],
        ),
        migrations.CreateModel(
            name="SubscriptionPlan",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50, unique=True)),
                (
                    "price_per_month",
                    models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True),
                ),
                (
                    "description",
                    models.TextField(blank=True, max_length=500, null=True),
                ),
                (
                    "stripe_product_id",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "stripe_price_id",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserSubscription",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4, unique=True)),
                (
                    "stripe_subscription_id",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "custom_subscription_price_per_month",
                    models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True),
                ),
                ("start_date", models.DateTimeField(auto_now_add=True)),
                ("end_date", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from datetime import datetime, timedelta, date

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.hashers import make_password
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.shortcuts import redirect
from django.urls import reverse, resolve, NoReverseMatch
from django.utils import timezone
from django.utils.http import url_has_allowed_host_and_scheme

from core.models import User, PasswordSecret
from core.models import RandomCode
from core.types.htmx import HtmxHttpRequest


def msg_if_valid_email_then_sent(request):
    return messages.success(
        request,
        f"<strong>If this is a valid email address</strong> then we have sent you an email.\n Please check spam, if you cannot find the email press forgot password again.",
    )


def set_password_generate(request: HtmxHttpRequest):
    if not request.user.is_superuser or not request.user.is_staff:
        return redirect("dashboard")

    USER = request.GET.get("id")
    NEXT = request.GET.get("next") or "index"

    if USER is None or not USER.isnumeric():
        messages.error(request, "User ID must be a valid integer")
        return redirect("dashboard")

    USER_OBJ = User.objects.filter(id=USER).first()

    if not USER_OBJ:
        messages.error(request, f"User not found")
        return redirect("dashboard")
    CODE = RandomCode(40)
    HASHED_CODE = make_password(CODE, salt=settings.SECRET_KEY)

    PWD_SECRET, created = PasswordSecret.objects.update_or_create(
        user=USER_OBJ,
        defaults={"expires": date.today() + timedelta(days=3), "secret": HASHED_CODE},
    )
    PWD_SECRET.save()
    messages.error(
        request,
        f'Successfully created a code. <a href="{reverse("user set password", args=(CODE,))}">{CODE}</a>',
    )

    if url_has_allowed_host_and_scheme(NEXT, allowed_hosts=None):
        try:
            resolve(NEXT)
            return redirect(NEXT)
        except NoReverseMatch:
            return redirect("dashboard")
    else:
        return redirect("dashboard")


def password_reset(request: HtmxHttpRequest):
    EMAIL = request.POST.get("email")

    # if not EMAIL_SERVER_ENABLED:
    #     messages.error(request, "Unfortunately our email server is not currently available.")

    if not EMAIL:
        msg_if_valid_email_then_sent(request)
        return redirect("login forgot_password")

    try:
        validate_email(EMAIL)
    except ValidationError as e:
        msg_if_valid_email_then_sent(request)
        return redirect("login forgot_password")

    USER = User.objects.filter(email=EMAIL).first()

    if not USER:
        msg_if_valid_email_then_sent(request)
        return redirect("login forgot_password")

    PasswordSecret.objects.filter(user=USER).all().delete()

    CODE = RandomCode(40)
    HASHED_CODE = make_password(CODE)
    expires_date = date.today() + timedelta(days=3)
    expires_datetime = timezone.make_aware(datetime.combine(expires_date, datetime.min.time()))

    PasswordSecret.objects.create(user=USER, expires=expires_datetime, secret=HASHED_CODE)

    # SEND_SENDGRID_EMAIL(USER.email, "Password Reset" ,f"""
    #         My Finances | Password Reset
    #         You've now got a new password reset code.
    #
    #         Reset Here: {request.build_absolute_uri(reverse('user set password', args=(CODE,)))}
    # """, request=request)
    print(f"code is {CODE}")

    msg_if_valid_email_then_sent(request)

    return redirect("login forgot_password")



import stripe
from typing import Union
from django.utils import timezone

from core.models import Organization
from billing.models import UserSubscription, SubscriptionPlan
from django.contrib.auth.models import User

from billing.service.entitlements import update_user_entitlements


def create_subscription(owner: Union[User, Organization], subscription_plan: SubscriptionPlan) -> UserSubscription:
    """
    Creates a new Stripe subscription for a user or organization.

    Args:
        owner: The user or organization subscribing.
        subscription_plan: The plan the owner is subscribing to.

    Returns:
        A UserSubscription object representing the subscription.
    """
    # Create a new Stripe subscription for the given owner (user or organization)
    stripe_subscription = stripe.Subscription.create(
        customer=owner.stripe_customer_id,
        items=[{"price": subscription_plan.stripe_price_id}],
    )

    # Create the corresponding UserSubscription record in your database
    user_subscription = UserSubscription.objects.create(
        owner=owner, subscription_plan=subscription_plan, stripe_subscription_id=stripe_subscription.id, start_date=timezone.now()
    )

    # Update user entitlements via Stripe entitlements
    update_user_entitlements(owner)

    return user_subscription


def cancel_subscription(user_subscription: UserSubscription) -> UserSubscription:
    """
    Cancels an active Stripe subscription and updates the local subscription record.

    Args:
        user_subscription: The subscription to cancel.

    Returns:
        The updated UserSubscription object with the end_date set.
    """
    stripe.Subscription.delete(subscription=user_subscription.stripe_subscription_id)

    # Mark the subscription as canceled in your local database
    user_subscription.end_date = timezone.now()
    user_subscription.save()

    # Update entitlements after cancellation
    update_user_entitlements(user_subscription.owner)

    return user_subscription


def handle_plan_change(user_subscription: UserSubscription, new_plan: SubscriptionPlan) -> UserSubscription:
    """
    Updates the user's Stripe subscription to a new plan, with proration handled automatically.

    Args:
        user_subscription: The current UserSubscription to update.
        new_plan: The new SubscriptionPlan to switch to.

    Returns:
        The updated UserSubscription object with the new plan.
    """
    stripe.Subscription.modify(
        user_subscription.stripe_subscription_id,
        cancel_at_period_end=False,  # Cancels the current plan immediately
        items=[{"price": new_plan.stripe_price_id}],
    )

    # Update the local subscription model
    user_subscription.subscription_plan = new_plan
    user_subscription.save()

    # Update entitlements after the plan change
    update_user_entitlements(user_subscription.owner)

    return user_subscription



# Generated by Django 5.1.4 on 2024-12-21 22:26

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("billing", "0001_initial"),
        ("core", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="billingusage",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.organization",
            ),
        ),
        migrations.AddField(
            model_name="billingusage",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="planfeature",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="features",
                to="billing.planfeaturegroup",
            ),
        ),
        migrations.AddField(
            model_name="stripecheckoutsession",
            name="features",
            field=models.ManyToManyField(related_name="checkout_sessions", to="billing.planfeature"),
        ),
        migrations.AddField(
            model_name="stripecheckoutsession",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.organization",
            ),
        ),
        migrations.AddField(
            model_name="stripecheckoutsession",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="stripecheckoutsession",
            name="plan",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="checkout_sessions",
                to="billing.subscriptionplan",
            ),
        ),
        migrations.AddField(
            model_name="planfeature",
            name="subscription_plan",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="features",
                to="billing.subscriptionplan",
            ),
        ),
        migrations.AddField(
            model_name="usersubscription",
            name="organization",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.organization",
            ),
        ),
        migrations.AddField(
            model_name="usersubscription",
            name="subscription_plan",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="billing.subscriptionplan",
            ),
        ),
        migrations.AddField(
            model_name="usersubscription",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddConstraint(
            model_name="billingusage",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    models.Q(("organization__isnull", False), ("user__isnull", True)),
                    models.Q(("organization__isnull", True), ("user__isnull", False)),
                    _connector="OR",
                ),
                name="billing_billingusage_check_user_or_organization",
            ),
        ),
        migrations.AddConstraint(
            model_name="stripecheckoutsession",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    models.Q(("organization__isnull", False), ("user__isnull", True)),
                    models.Q(("organization__isnull", True), ("user__isnull", False)),
                    _connector="OR",
                ),
                name="billing_stripecheckoutsession_check_user_or_organization",
            ),
        ),
        migrations.AddConstraint(
            model_name="usersubscription",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    models.Q(("organization__isnull", False), ("user__isnull", True)),
                    models.Q(("organization__isnull", True), ("user__isnull", False)),
                    _connector="OR",
                ),
                name="billing_usersubscription_check_user_or_organization",
            ),
        ),
    ]



from django.core.management.base import BaseCommand
from core.service.maintenance.expire.run import expire_and_cleanup_objects


class Command(BaseCommand):
    """
    Runs automation scripts to make sure objects are up to date, expired objects are deleted, etc.
    """

    def handle(self, *args, **kwargs):
        self.stdout.write("Running expire + cleanup script...")
        self.stdout.write(expire_and_cleanup_objects())



from __future__ import annotations

import logging

from django.db.models.signals import post_migrate
from django.dispatch import receiver

from core.data.default_feature_flags import default_feature_flags
from core.models import FeatureFlags


@receiver(post_migrate)
def update_feature_flags(**kwargs):
    for feature_flag in default_feature_flags:
        existing_item = FeatureFlags.objects.filter(name=feature_flag.name).first()

        if existing_item:
            name, value, description = (
                existing_item.name,
                existing_item.value,
                existing_item.description,
            )

            existing_item.name = name
            existing_item.description = description

            if existing_item.name != name or existing_item.description != description:
                existing_item.save()
                logging.info(f"Updated feature flag: {feature_flag.name}")
        else:
            FeatureFlags.objects.create(name=feature_flag.name, value=feature_flag.default, description=feature_flag.description)
            logging.info(f"Added feature flag: {feature_flag.name}")


#
# @receiver(post_migrate)
# def update_quota_limits(**kwargs):
#     for group in default_quota_limits:
#         for item in group.items:
#             existing = QuotaLimit.objects.filter(slug=f"{group.name}-{item.slug}").first()
#             if existing:
#                 name, value, adjustable, description, limit_type = (
#                     existing.name,
#                     existing.value,
#                     existing.adjustable,
#                     existing.description,
#                     existing.limit_type,
#                 )
#                 existing.name = item.name
#                 existing.adjustable = item.adjustable
#                 existing.description = item.description
#                 existing.limit_type = item.period
#                 if (
#                     item.name != name
#                     or item.default_value != value
#                     or item.adjustable != adjustable
#                     or item.description != description
#                     or item.period != limit_type
#                 ):
#                     logging.info(f"Updated QuotaLimit {item.name}")
#                     existing.save()
#             else:
#                 QuotaLimit.objects.create(
#                     name=item.name,
#                     slug=f"{group.name}-{item.slug}",
#                     value=item.default_value,
#                     adjustable=item.adjustable,
#                     description=item.description,
#                     limit_type=item.period,
#                 )
#                 logging.info(f"Added QuotaLimit {item.name}")



from django.urls import path

from . import kick, switch_team, invites, leave, create, edit_permissions
from .create_user import create_user_endpoint

urlpatterns = [
    path("edit_permissions/", edit_permissions.edit_user_permissions_endpoint, name="edit_permissions"),
    path(
        "kick/<int:user_id>",
        kick.kick_user,
        name="kick",
    ),
    path(
        "switch_team/<int:team_id>/",
        switch_team.switch_team,
        name="switch_team",
    ),
    path(
        "switch_team/",
        switch_team.switch_team,
        name="switch_team input",
    ),
    path(
        "create_user/",
        create_user_endpoint,
        name="create_user",
    ),
    # INVITES #
    path(
        "invite/",
        invites.send_user_team_invite,
        name="invite",
    ),
    path(
        "join/<str:code>/accept/",
        invites.accept_team_invite,
        name="join accept",
    ),
    path(
        "join/<str:code>/decline/",
        invites.decline_team_invite,
        name="join decline",
    ),
    # LEAVE TEAM #
    path(
        "leave/<int:team_id>/confirm/",
        leave.leave_team_confirmed,
        name="leave confirm",
    ),
    path(
        "create/",
        create.create_team,
        name="create",
    ),
    path("get_dropdown/", switch_team.get_dropdown, name="get_dropdown"),
]

app_name = "teams"



import botocore.client
from botocore.config import Config
from botocore.exceptions import NoCredentialsError, PartialCredentialsError

from core.utils.settings_helper import get_var

import boto3
import logging

logger = logging.getLogger(__name__)

DEBUG_LEVEL = get_var("AWS_LOG_LEVEL", default="debug")
DEBUG_LEVEL = "debug" if DEBUG_LEVEL == "debug" else "info" if DEBUG_LEVEL == "info" else None


class Boto3Handler:
    def __init__(self):
        self.initiated: bool = False
        self.region_name: str = get_var("AWS_REGION_NAME", "eu-west-2")
        self.aws_access_key_id: str = get_var("AWS_ACCESS_KEY_ID")
        self.aws_access_key_secret: str = get_var("AWS_ACCESS_KEY")
        self.scheduler_lambda_arn: str = get_var("SCHEDULER_LAMBDA_ARN")
        self.scheduler_lambda_access_role_arn: str = get_var("SCHEDULER_LAMBDA_ACCESS_ROLE_ARN")
        self.scheduler_invoices_group_name: str = get_var("SCHEDULER_INVOICES_GROUP_NAME")
        self.dynamodb_client = None
        self.scheduler_client = None

        print(f"Region: {self.region_name}")
        print("| has aws access key id" if self.aws_access_key_id else "X no aws access key id")
        print("| has aws access key secret" if self.aws_access_key_secret else "X no aws access key secret")
        print("| has scheduler lambda arn" if self.scheduler_lambda_arn else "X no scheduler lambda arn")
        print(
            "| has scheduler lambda access role arn" if self.scheduler_lambda_access_role_arn else "X no scheduler lambda access role arn"
        )
        print("| has scheduler invoices group name" if self.scheduler_invoices_group_name else "X no scheduler invoices group name")

        self._initiate_clients()

    def _initiate_session(self):
        self._boto3_config = Config(region_name=self.region_name, signature_version="v4", retries={"max_attempts": 10, "mode": "standard"})

        self._boto3_session = boto3.Session(
            # aws_access_key_id=self.aws_access_key_id,
            # aws_secret_access_key=self.aws_access_key_secret,
            region_name=self.region_name
        )

        if DEBUG_LEVEL == "debug":
            boto3.set_stream_logger("", level=logging.DEBUG)
        else:
            boto3.set_stream_logger("", level=logging.INFO)

    def _initiate_clients(self):
        if get_var("AWS_DISABLED", "").lower() == "true":
            logger.info("The variable AWS_DISABLED is present, not initiating boto3")
            return

        if not get_var("AWS_ENABLED"):
            logger.error("The variable AWS_ENABLED is not present, not initiating boto3")
            return

        self._initiate_session()

        try:
            if not self._boto3_session.client("sts").get_caller_identity():
                logger.info("No AWS Credentials found, not initiating clients.")
                return
        except (NoCredentialsError, PartialCredentialsError) as error:
            logger.error(error)
            return None

        self._schedule_client = self._boto3_session.client("scheduler")
        self.schedule_client = self._schedule_client
        self._dynamodb_client = self._boto3_session.client("dynamodb")
        self.dynamodb_client = self._dynamodb_client

        self.SCHEDULE_EXCEPTIONS = self._schedule_client.exceptions
        self.DYNAMO_EXCEPTIONS = self._dynamodb_client.exceptions
        self.initiated = True

        logger.info("Boto3Handler has been initiated!")


BOTO3_HANDLER = Boto3Handler()



import datetime
import functools
import logging

from rest_framework.response import Response

logger = logging.getLogger(__name__)


# add type hints for these deprecation dates


def deprecated(deprecation_date: datetime.datetime | None = None, end_of_life_date: datetime.datetime | None = None):
    """
    Returns a decorator which informs requester that the decorated endpoint has been deprecated.
    """

    def decorator_deprecated(func):
        """Amend the request with information that the endpoint has been deprecated and when it will be removed"""

        @functools.wraps(func)
        def wrapper_deprecated(*args, **kwargs):
            # do something before handling the request, could e.g. issue a django signal
            logger.warning("Deprecated endpoint %s called", func.__name__)

            if end_of_life_date and datetime.datetime.now() > end_of_life_date:
                return Response(
                    {"success": False, "message": "This endpoint is no longer available"},
                    status=410,
                    headers={"X-Deprecated": "", "X-Deprecation-Date": deprecation_date, "X-End-Of-Life-Date": end_of_life_date},
                )

            response: Response = func(*args, **kwargs)

            # amend the response with deprecation information
            if isinstance(response, Response):
                response.headers["X-Deprecated"] = ""
                if deprecation_date:
                    response.headers["X-Deprecation-Date"] = deprecation_date
                if end_of_life_date:
                    response.headers["X-End-Of-Life-Date"] = deprecation_date
            return response

        return wrapper_deprecated

    return decorator_deprecated



from django.urls import path

from billing.views.dashboard import billing_dashboard_endpoint, all_subscriptions_htmx_endpoint
from billing.views.stripe_misc import customer_client_portal_endpoint
from billing.views.stripe_webhooks import stripe_listener_webhook_endpoint

from .views.return_urls.success import stripe_success_return_endpoint
from .views.change_plan import change_plan_endpoint

urlpatterns = [
    path("dashboard/billing/", billing_dashboard_endpoint, name="dashboard"),
    path("api/billing/all_subscriptions/", all_subscriptions_htmx_endpoint, name="all_subscriptions"),
    # path("dashboard/billing/", RedirectView.as_view(url="/dashboard"), name="dashboard"),
    path("api/public/webhooks/receive/payments/stripe/", stripe_listener_webhook_endpoint, name="receive_stripe_webhook"),
    path("api/billing/stripe/change_plan/", change_plan_endpoint, name="change_plan"),
    path("dashboard/billing/stripe/portal/", customer_client_portal_endpoint, name="stripe_customer_portal"),
    path("dashboard/billing/stripe/checkout/response/success/", stripe_success_return_endpoint, name="stripe_checkout_success_response"),
    path("dashboard/billing/stripe/checkout/response/failed/", stripe_success_return_endpoint, name="stripe_checkout_failed_response"),
]

app_name = "billing"



from typing import Optional

from django.db.models import When, Case, BooleanField, QuerySet
from django.shortcuts import render

from core.models import Organization, User
from core.service.teams.fetch import get_all_users_teams
from core.types.requests import WebRequest


def teams_dashboard(request: WebRequest):
    context: dict[str, str | int] = {}

    users_team: Optional[Organization] = request.user.logged_in_as_team

    if not users_team:
        user_with_counts = User.objects.prefetch_related("teams_joined", "teams_leader_of").get(pk=request.user.pk)
        return render(
            request,
            "pages/settings/teams/main.html",
            context
            | {
                "team": None,
                "team_count": user_with_counts.teams_joined.count() + user_with_counts.teams_leader_of.count(),
            },
        )

    try:
        team = (
            Organization.objects.annotate(
                is_leader=Case(
                    When(leader=request.user, then=True),
                    default=False,
                    output_field=BooleanField(),
                ),
            )
            .prefetch_related("members", "permissions")
            .get(id=users_team.id)
        )

        user_permissions: dict[User, list] = {}

        for member in team.members.all():
            member_perms = list(team.permissions.filter(user=member).values_list("scopes", flat=True))

            if len(member_perms) > 0:
                user_permissions[member] = member_perms[0]
            else:
                user_permissions[member] = []

    except Organization.DoesNotExist:
        user_with_counts = User.objects.prefetch_related("teams_joined", "teams_leader_of").get(pk=request.user.pk)
        return render(
            request,
            "pages/settings/teams/main.html",
            context
            | {
                "team": None,
                "team_count": user_with_counts.teams_joined.count() + user_with_counts.teams_leader_of.count(),
            },
        )

    return render(
        request,
        "pages/settings/teams/main.html",
        context | {"team": team, "user_permissions": user_permissions},
    )


def login_to_team_page(request: WebRequest, all_teams: QuerySet[Organization]):
    print(all_teams)
    return render(request, "pages/settings/teams/login_to_team.html", {"team_list": all_teams})


def teams_dashboard_handler(request: WebRequest):
    all_teams: QuerySet[Organization] = get_all_users_teams(request)
    logged_in_team: Organization | None = request.user.logged_in_as_team

    if not logged_in_team:
        return login_to_team_page(request, all_teams)
    return teams_dashboard(request)



from core.models import User, Organization


def get_actor_from_stripe_customer(stripe_customer_id: str) -> User | Organization | None:
    return (
        User.objects.filter(stripe_customer_id=stripe_customer_id).first()
        or Organization.objects.filter(stripe_customer_id=stripe_customer_id).first()
    )



import json
import os
from typing import TypedDict

from django.core.management.base import BaseCommand
from django.utils.termcolors import colorize


class ContributorsItem(TypedDict):
    name: str
    username: str
    role: str
    tags: list[str]


class Command(BaseCommand):
    """
    Adds contributors HTML table to README.md file.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.script_dir = os.path.dirname(os.path.abspath(__file__))
        self.contributors_json_path = os.path.join(self.script_dir, "contributors.json")
        self.readme_path = "README.md"

    def add_arguments(self, parser):
        parser.add_argument("action", type=str, help="sync, list, add, edit")

        parser.add_argument("--sort", type=str, help="Sort by: name, username or role")
        parser.add_argument("--limit", type=int, default=20, help="Limit results by amount")

        parser.add_argument("name", type=str, nargs="?", help="users human/readable name")
        parser.add_argument("--name", type=str, help="users human/readable name")
        parser.add_argument("username", type=str, nargs="?", help="github username")
        parser.add_argument("--username", type=str, help="github username")
        parser.add_argument("role", type=str, nargs="?", help="role in team")
        parser.add_argument("--role", type=str, help="role in team")
        parser.add_argument("tags", type=str, nargs="*", help="comma separated list of tags")
        parser.add_argument("--tags", type=str, help="comma separated list of tags")

    def handle(self, *args, **kwargs):
        action = kwargs["action"]

        if action == "sync":
            self.sync_contributors()
        elif action == "list":
            self.list_contributors(*args, **kwargs)
        elif action == "add":
            self.add_contributor(*args, **kwargs)
        elif action == "help":
            self.stdout.write(
                colorize(
                    """
Please provide valid action.
- sync
- list --sort <name|username|role> --limit <int>
- add <name> <username> <role> <tags>
    to use multi-word usernames or names, surround with quotes
    tags can be space separated
    to change the order use --, e.g. --name bob
            """,
                    fg="red",
                    opts=("bold",),
                )
            )
        else:
            self.stdout.write(colorize("Please provide valid action. \n - sync \n - list \n - add \n - help", fg="red", opts=("bold",)))

    def add_contributor(self, *args, **kwargs):
        name: str = kwargs.get("name")
        username: str = kwargs.get("username")
        role: str = kwargs.get("role")
        tags: list[str] = kwargs.get("tags")

        if not name or not username or not role or not tags:
            return self.stdout.write(colorize("Please provide name, username, role and tags", fg="red", opts=("bold",)))

        if not all([t in ["👑", "🖥", "🎨", "📖", "🐳", "♻", "🐞"] for t in tags]):
            return self.stdout.write(
                colorize(f"Please provide valid tags. Valid tags are: 👑, 🖥, 🎨, 📖, 🐳, ♻, 🐞", fg="red", opts=("bold",))
            )

        contributors_data: list[ContributorsItem] = self._read_contributor_file()

        for user in contributors_data:
            if user["username"] == username:
                return self.stdout.write(colorize("User already exists", fg="red", opts=("bold",)))

        contributor_obj = ContributorsItem(name=name, username=username, role=role, tags=tags)

        if contributors_data:
            contributors_data.append(contributor_obj)
        else:
            return self.stdout.write(
                colorize("contributors.json file not found. Please make sure the file exists.", fg="red", opts=("bold",))
            )

        self._save_contributors_file(contributors_data)

    def list_contributors(self, *args, **kwargs):
        contributors_data: list[ContributorsItem] | None = self._read_contributor_file()

        if not contributors_data:
            return

        if kwargs.get("sort") == "name":
            contributors_data = sorted(contributors_data, key=lambda d: d.get("name", ""))
        elif kwargs.get("sort") == "username":
            contributors_data = sorted(contributors_data, key=lambda d: d.get("username", ""))
        elif kwargs.get("sort") == "role":
            contributors_data = sorted(contributors_data, key=lambda d: d.get("role", ""))

        if limit := kwargs.get("limit"):
            bef_count = len(contributors_data)
            contributors_data = contributors_data[:limit]
            aft_count = len(contributors_data)

        max_w_name = max(len(d.get("name", "")) + 4 for d in contributors_data)
        max_w_username = max(len(d.get("username", "")) + 4 for d in contributors_data)
        max_w_role = max(len(d.get("role", "")) + 4 for d in contributors_data)

        row_str = "{:<{max_w_name}} {:<{max_w_username}} {:<{max_w_role}} {:<10}"

        header = row_str.format(
            "Name", "Username", "Role", "Tags", max_w_name=max_w_name, max_w_username=max_w_username, max_w_role=max_w_role
        )

        self.stdout.write(header)

        for user in contributors_data:
            row = row_str.format(
                user.get("name"),
                user.get("username"),
                user.get("role"),
                " ".join(user.get("tags")),
                max_w_name=max_w_name,
                max_w_username=max_w_username,
                max_w_role=max_w_role,
            )
            self.stdout.write(row)

        if limit:
            # noinspection PyUnboundLocalVariable
            self.stdout.write(f"\nShowing {aft_count} of {bef_count} contributors\n")

    def sync_contributors(self):
        contributors_data = self._read_contributor_file()

        if not contributors_data:
            return

        # Path to contributors.json file in the same directory as the script

        # HTML template for each contributor entry
        contributor_template = """
                <td align="center">
                    <a href="https://github.com/{username}" title="{username}">
                        <img title="{role}" src="https://github.com/{username}.png" width="100px;" alt="" />
                        <br />
                        <sub>
                            <b>{name}</b>
                        </sub>
                    </a>
                    <br />
                    {tags}
                </td>
                """

        # Function to generate HTML for a contributor
        def generate_contributor_html(contributor):
            tags_html = "".join(
                [
                    f'<a href="{tag_link}" title="{tag_title}">{tag_icon}</a>'
                    for tag_icon, tag_link, tag_title in [
                        (
                            "👑",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=user%3A{contributor["username"]}',
                            "Project Lead",
                        ),
                        (
                            "🖥",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Backend",
                        ),
                        (
                            "📖",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Documentation",
                        ),
                        (
                            "🎨",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Frontend",
                        ),
                        (
                            "🐞",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Bug Fixes",
                        ),
                        (
                            "🧪",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Added Tests",
                        ),
                        (
                            "🐳",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Docker",
                        ),
                        (
                            "♻",
                            f'https://github.com/TreyWW/MyFinances/pulls?q=is%3Apr+author%3A{contributor["username"]}',
                            "Refactored Files",
                        ),
                    ]
                    if tag_icon in contributor.get("tags", [])
                ]
            )
            return contributor_template.format(
                username=contributor["username"],
                name=contributor["name"],
                role=contributor.get("role"),
                tags=tags_html,
            )

        # Generate HTML for contributors
        contributors_html = ""
        users_per_row = 5
        for index, contributor in enumerate(contributors_data, start=1):
            if (index - 1) % users_per_row == 0 and index > 1:
                contributors_html += "</tr><tr>"
            contributors_html += f"<td>{generate_contributor_html(contributor)}</td>"

        # Wrap the entire content in a table row
        contributors_html = f"<tr>{contributors_html}</tr>"

        try:
            # Read the README.md file
            with open(self.readme_path, encoding="utf-8") as readme_file:
                readme_content = readme_file.read()
        except FileNotFoundError:
            self.stderr.write(self.style.ERROR("README.md file not found. Please make sure the file exists."))
            return

        # Insert the generated HTML between the comments
        start_comment = "<!-- CONTRIBUTORS TABLE START -->"
        end_comment = "<!-- CONTRIBUTORS TABLE END -->"
        start_index = readme_content.find(start_comment) + len(start_comment)
        end_index = readme_content.find(end_comment)

        new_readme_content = readme_content[:start_index] + "\n<table>\n" + contributors_html + "</table>\n" + readme_content[end_index:]

        self._save_readme_file(new_readme_content)

        self.stdout.write(self.style.SUCCESS("HTML table inserted into README.md successfully."))

    def _read_contributor_file(self) -> list[ContributorsItem] | None:
        try:
            # Load JSON data from contributors.json file
            with open(self.contributors_json_path, encoding="utf-8") as json_file:
                contributors_json = json_file.read()
                return json.loads(contributors_json)
        except FileNotFoundError:
            self.stderr.write(self.style.ERROR("contributors.json file not found. Please make sure the file exists."))
            return None
        except json.JSONDecodeError:
            self.stderr.write(self.style.ERROR("Error decoding JSON data from contributors.json file. Please check the file contents."))
            return None

    def _read_readme_file(self) -> str | None:
        try:
            # Read the README.md file
            with open(self.readme_path, encoding="utf-8") as readme_file:
                return readme_file.read()
        except FileNotFoundError:
            self.stderr.write(self.style.ERROR("README.md file not found. Please make sure the file exists."))
            return None

    def _save_contributors_file(self, contributors_data: list[ContributorsItem]):
        try:
            # Save JSON data to contributors.json file
            with open(self.contributors_json_path, "w", encoding="utf-8") as json_file:
                json.dump(contributors_data, json_file, indent=4, ensure_ascii=False)
        except FileNotFoundError:
            self.stderr.write(self.style.ERROR("contributors.json file not found. Please make sure the file exists."))
            return
        except json.JSONDecodeError:
            self.stderr.write(self.style.ERROR("Error encoding JSON data to contributors.json file. Please check the file contents."))
            return

    def _save_readme_file(self, new_readme_content):
        with open(self.readme_path, "w", encoding="utf-8") as readme_file:
            readme_file.write(new_readme_content)



from __future__ import annotations

from django.urls import path

from . import fetch, status

# from . import send

urlpatterns = [
    # path(
    #     "send/single/",
    #     send.send_single_email_view,
    #     name="send single",
    # ),
    # path(
    #     "send/bulk/",
    #     send.send_bulk_email_view,
    #     name="send bulk",
    # ),
    path("fetch/", fetch.fetch_all_emails, name="fetch"),
    path("get_status/<str:status_id>/", status.get_status_view, name="get_status"),
    path("refresh_statuses/", status.refresh_all_statuses_view, name="refresh statuses"),
]

app_name = "emails"



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render

from core.models import Organization
from core.types.htmx import HtmxHttpRequest


def return_error_notif(request: HtmxHttpRequest, message: str):
    messages.error(request, message)
    resp = render(request, "partials/messages_list.html", status=200)
    resp["HX-Trigger-After-Swap"] = "leave_team_error"
    return resp


def leave_team_confirmed(request: HtmxHttpRequest, team_id):
    team: Organization | None = Organization.objects.filter(id=team_id).first()

    if not team:
        return return_error_notif(request, "Team not found")

    if team.leader == request.user:  # may be changed in the future. If no members allow delete
        return return_error_notif(request, "You cannot leave your own team")

    if request.user.teams_joined.filter(id=team_id).exists():
        team.members.remove(request.user)
        messages.success(request, f"You have successfully left the team {team.name}")
        response = HttpResponse(status=200)
        response["HX-Refresh"] = "true"
        return response
    else:
        return return_error_notif(request, "You are not a member of this team")



from textwrap import dedent

import django_ratelimit
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login, logout, authenticate
from django.contrib.auth.hashers import check_password
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render, redirect
from django.urls import resolve, reverse
from django.urls.exceptions import Resolver404
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.http import require_GET, require_POST
from django_ratelimit.decorators import ratelimit

from core.decorators import not_authenticated
from core.types.htmx import HtmxAnyHttpRequest

from core.utils.settings_helper import send_email, ARE_EMAILS_ENABLED
from core.models import LoginLog, User, VerificationCodes, AuditLog
from core.views.auth.verify import create_magic_link


@require_GET
@not_authenticated
def login_initial_page(request: HttpRequest):
    redirect_url = request.GET.get("next")

    return render(
        request,
        "core/auth/login_initial.html",
        {
            "github_enabled": getattr(settings, "SOCIAL_AUTH_GITHUB_ENABLED", False),
            "google_enabled": getattr(settings, "SOCIAL_AUTH_GOOGLE_OAUTH2_ENABLED", False),
            "next": redirect_url,
        },
    )


@not_authenticated
@require_POST
def login_manual(request: HttpRequest):
    email = request.POST.get("email")
    password = request.POST.get("password")
    redirect_url = request.POST.get("next", "")

    if not email:
        messages.error(request, "Please enter an email")
        return redirect_to_login("", redirect_url)

    try:
        validate_email(email)
    except ValidationError:
        messages.error(request, "Please enter a valid email")
        return redirect_to_login("", redirect_url)

    if not password:
        messages.error(request, "Please enter a password")
        return redirect_to_login(email, redirect_url)

    user = authenticate(request, username=email, password=password)

    if not user:
        messages.error(request, "Incorrect email or password")
        return redirect_to_login(email, redirect_url)

    if user.awaiting_email_verification and ARE_EMAILS_ENABLED:  # type: ignore[attr-defined]
        messages.error(request, "You must verify your email before logging in.")
        return redirect_to_login(email, redirect_url)

    login(request, user)

    if user.require_change_password:  # type: ignore[attr-defined]
        messages.warning(request, "You have been requested by an administrator to change your account password.")
        return redirect("core:settings:change_password")

    if url_has_allowed_host_and_scheme(redirect_url, allowed_hosts=None):
        try:
            resolve(redirect_url)
            return redirect(redirect_url)
        except Resolver404:
            return redirect("dashboard")
    else:
        return redirect("dashboard")


def redirect_to_login(email: str, redirect_url: str):
    if not url_has_allowed_host_and_scheme(redirect_url, allowed_hosts=None):
        redirect_url = reverse("dashboard")
    return redirect(f"{reverse('core:auth:login')}?email={email}&next={redirect_url}")


def render_error_toast_message(request: HttpRequest, message: str) -> HttpResponse:
    messages.error(request, message)
    return render_toast_message(request)


def render_toast_message(request: HttpRequest) -> HttpResponse:
    return render(request, "base/toasts.html")  # htmx will handle the toast


class MagicLinkRequestView(View):
    @method_decorator(ratelimit(key="post:email", method=django_ratelimit.UNSAFE, rate="5/m"))
    @method_decorator(ratelimit(key="post:email", method=django_ratelimit.UNSAFE, rate="10/5m"))
    @method_decorator(ratelimit(key="ip", method=django_ratelimit.UNSAFE, rate="2/m"))
    @method_decorator(ratelimit(key="ip", method=django_ratelimit.UNSAFE, rate="3/10m"))
    @method_decorator(ratelimit(key="ip", method=django_ratelimit.UNSAFE, rate="6/1h"))
    def post(self, request: HtmxAnyHttpRequest) -> HttpResponse | bool:
        if request.user.is_authenticated:
            return redirect("dashboard")
        if not request.htmx:
            return redirect("core:auth:login")

        email = request.POST.get("email")
        try:
            user = User.objects.get(email=email)
        except User.DoesNotExist:
            return self.send_message(request)

        if not user.is_active:
            return self.send_message(request, "This account is not currently active.", False)

        magic_link, plain_token = create_magic_link(user, service="login")
        self.send_magic_link_email(request, user, str(magic_link.uuid), plain_token)
        self.send_message(request, should_redirect=False)
        return render(request, "core/auth/magic_link_waiting.html", {"email": request.POST.get("email")})

    def send_message(
        self, request: HttpRequest, message: str = "", success: bool = True, should_redirect: bool = True
    ) -> HttpResponse | bool:
        message = message or "If this is a valid email address, we have sent you an email! Keep this tab open!"
        if success:
            messages.success(request, message)
        else:
            messages.error(request, message)

        if should_redirect:
            return render_toast_message(request)
        else:
            return True

    def send_magic_link_email(self, request: HttpRequest, user: User, uuid: str, plain_token: str) -> None:
        magic_link_url = request.build_absolute_uri(
            reverse("core:auth:login magic_link verify", kwargs={"uuid": uuid, "token": plain_token})
        )

        send_email(
            destination=user.email,
            subject="Login Request",
            content=dedent(
                f"""
            Hi {user.first_name if user.first_name else "User"},

            A login request was made on your MyFinances account. If this was not you, please ignore
            this email.

            If you would like to login, please use the following link: \n {magic_link_url}
        """
            ),
        )


class MagicLinkWaitingView(View):
    def post(self, request: HtmxAnyHttpRequest) -> HttpResponse:
        if request.user.is_authenticated:
            return redirect("dashboard")
        if not request.htmx:
            return redirect("core:auth:login")
        return render(request, "core/auth/magic_link_waiting.html", {"email": request.POST.get("email")})


class MagicLinkVerifyView(View):
    def get(self, request: HttpRequest, uuid: str, token: str) -> HttpResponse:
        if request.user.is_authenticated:
            return redirect("dashboard")

        magic_link = get_magic_link(uuid)

        magic_link_valid, magic_link_msg = is_magiclink_valid(magic_link, token)
        if not magic_link_valid:
            messages.error(request, magic_link_msg)
            return redirect("core:auth:login")

        # user = magic_link.user
        # magic_link.delete()
        # user.backend = "backend.auth_backends.EmailInsteadOfUsernameBackend"
        # login(request, magic_link.user)

        return render(request, "core/auth/magic_link_verify.html", {"uuid": uuid, "token": token})

        #
        # messages.success(request, "Successfully logged in")
        # # TODO: Add page to make sure they click an extra "verify request btn"
        # return redirect("dashboard")


class MagicLinkVerifyDecline(View):
    def post(self, request: HtmxAnyHttpRequest, uuid: str, token: str) -> HttpResponse:
        if request.user.is_authenticated or not request.htmx:
            return redirect("dashboard")

        magic_link = get_magic_link(uuid)
        magic_link_valid, magic_link_msg = is_magiclink_valid(magic_link, token)

        if not magic_link_valid or magic_link is None:
            messages.error(request, magic_link_msg)
            return render_toast_message(request)
        else:
            user = magic_link.user
            magic_link.delete()

            AuditLog.objects.create(user=user, action="magic link declined")
            messages.success(request, "Successfully declined the magic link verification request.")
            return render(request, "core/auth/_magic_link_partial.html", {"declined": True})


class MagicLinkVerifyAccept(View):
    def post(self, request: HtmxAnyHttpRequest, uuid: str, token: str) -> HttpResponse:
        if request.user.is_authenticated or not request.htmx:
            return redirect("dashboard")

        magic_link = get_magic_link(uuid)
        magic_link_valid, magic_link_msg = is_magiclink_valid(magic_link, token)

        if not magic_link_valid or magic_link is None:
            messages.error(request, magic_link_msg)
            return render_toast_message(request)
        else:
            user = magic_link.user
            magic_link.delete()

            user.backend = "backend.auth_backends.EmailInsteadOfUsernameBackend"  # type: ignore[attr-defined]
            LoginLog.objects.create(user=user, service=LoginLog.ServiceTypes.MAGIC_LINK)
            AuditLog.objects.create(user=user, action="magic link accepted")
            login(request, magic_link.user)

            messages.success(request, "Successfully accepted the magic link verification request.")
            return render(request, "core/auth/_magic_link_partial.html", {"accepted": True})


def is_magiclink_valid(magic_link: VerificationCodes | None, token: str) -> tuple[bool, str]:
    if not magic_link:
        return False, "Invalid magic link"

    if not magic_link.is_active():
        return False, "This link has expired"

    if not check_password(token, magic_link.token):
        return False, "Invalid magic link"

    return True, ""


def get_magic_link(uuid: str) -> VerificationCodes | None:
    try:
        return VerificationCodes.objects.get(uuid=uuid, service="login")
    except VerificationCodes.DoesNotExist:
        return None


def logout_view(request):
    logout(request)

    messages.success(request, "You've now been logged out.")

    return redirect("core:auth:login")  # + "?next=" + request.POST.get('next'))


@not_authenticated
def forgot_password_page(request: HttpRequest):
    if request.user.is_authenticated:
        return redirect("dashboard")
    return render(request, "core/auth/forgot_password.html")



from django.urls import path

from core.views.settings.view import change_password, view_settings_page_endpoint

urlpatterns = [
    path("", view_settings_page_endpoint, name="dashboard"),
    path("<str:page>/", view_settings_page_endpoint, name="dashboard with page"),
    path(
        "profile/change_password/",
        change_password,
        name="change_password",
    ),
]

app_name = "settings"



from core.api.public import APIAuthToken
from core.models import User, Organization


def get_api_key_by_name(owner: User | Organization, key_name: str) -> APIAuthToken | None:
    return APIAuthToken.filter_by_owner(owner).filter(name=key_name, active=True).first()


def get_api_key_by_id(owner: User | Organization, key_id: str | int) -> APIAuthToken | None:
    return APIAuthToken.filter_by_owner(owner).filter(id=key_id, active=True).first()



from __future__ import annotations

from django.http import HttpResponseBadRequest
from core.service.modals.registry import get_modal
from core.types.requests import WebRequest


def open_modal_endpoint(request: WebRequest, modal_name):
    modal_class = get_modal(modal_name)
    if not modal_class:
        print(f"Modal {modal_name} not found.")
        return HttpResponseBadRequest("Something went wrong with loading this modal.")
    modal_instance = modal_class()
    return modal_instance.get(request)



import logging

from django.views.decorators.http import require_http_methods
from django.contrib.auth import update_session_auth_hash
from django.contrib import messages
from django.shortcuts import redirect
from django.shortcuts import render

from core.config import CoreConfig
from core.types.requests import WebRequest
from core.models import TracebackError


@require_http_methods(["GET"])
def view_settings_page_endpoint(request: WebRequest, page: str | None = None):
    context = {}

    # If no page is provided, default to the 'profile' page
    if not page:
        template = "core/settings/pages/profile.html"
        # Check if 'on_main' query parameter is present to decide layout
        if not request.GET.get("on_main"):
            context["page_template"] = template
            return render(request, "core/settings/main.html", context)
        return render(request, template, context)

    # If a page is provided, look for its handler
    handler_path = CoreConfig().SETTINGS_PAGE_CONTEXT_HANDLERS.get(page)

    if not handler_path:
        logging.error(f"Settings page handler not found for {page}")
        messages.error(request, "Settings page not found")
        if request.htmx:
            return render(request, "core/base/toast.html")
        return redirect("core:settings:dashboard")

    # Extract the module path and function name from the handler path
    module_path, func_name = handler_path.rsplit(".", 1)

    try:
        # Dynamically import the module and get the handler function
        module = __import__(module_path, fromlist=[func_name])
        context_function = getattr(module, func_name)
        context_function(request, context)  # Call the handler function to fill the context
    except (ImportError, AttributeError) as e:
        print(e)
        messages.error(request, f"Settings page handler not implemented for {page}. Please contact the site support.")
        TracebackError.objects.create(user=request.user, error=f"Settings page handler not implemented for {page}")
        if request.htmx:
            return render(request, "core/base/toast.html")
        return redirect("core:settings:dashboard")

    # Default template for the provided page
    template = f"core/settings/pages/{page}.html"

    # Check if 'on_main' query parameter is present to decide layout
    if not request.GET.get("on_main"):
        context["page_template"] = template
        return render(request, "core/settings/main.html", context)

    # Otherwise, render the page directly
    response = render(request, template, context)
    response.no_retarget = True  # type: ignore[attr-defined]
    return response


def change_password(request: WebRequest):
    if request.method == "POST":
        current_password = request.POST.get("current_password")
        password = request.POST.get("password")
        confirm_password = request.POST.get("confirm_password")

        error = validate_password_change(request.user, current_password, password, confirm_password)

        if error:
            messages.error(request, error)
            return redirect("core:settings:change_password")

        # If no errors, update the password
        request.user.set_password(password)
        request.user.save()
        update_session_auth_hash(request, request.user)
        messages.success(request, "Successfully changed your password.")
        return redirect("core:settings:dashboard")

    return render(request, "pages/reset_password.html", {"type": "change"})


def validate_password_change(user, current_password, new_password, confirm_password):
    if not user.check_password(current_password):
        return "Incorrect current password"

    if new_password != confirm_password:
        return "Passwords don't match"

    if not new_password:
        return "Something went wrong, no password was provided."

    if len(new_password) < 8 or len(new_password) > 128:
        return "Password must be between 8 and 128 characters."

    return None



from core.models import User, Organization
import stripe


def get_or_create_customer_id(actor: User | Organization) -> str:
    if actor.stripe_customer_id:
        return actor.stripe_customer_id

    return create_stripe_customer_id(actor)


def create_stripe_customer_id(actor: User | Organization) -> str:
    if isinstance(actor, User):
        customer = stripe.Customer.create(
            email=actor.email,
            name=actor.get_full_name(),
        )
        actor.stripe_customer_id = customer.id
        actor.save()

        return customer.id

    else:
        customer = stripe.Customer.create(
            email=actor.leader.email,
            name=actor.name,
        )

        actor.stripe_customer_id = customer.id
        actor.save()

        return customer.id



import stripe

from core.utils.calendar import timezone_now
from billing.models import StripeCheckoutSession, StripeWebhookEvent, UserSubscription


def checkout_completed(webhook_event: StripeWebhookEvent):
    event_data: stripe.checkout.Session = webhook_event.data["object"]

    stripe_session_obj = StripeCheckoutSession.objects.filter(
        uuid=event_data.metadata.get("dj_checkout_uuid", "doesn't_exist") if event_data.metadata else "doesn't_exist"
    ).first()  # type: ignore[misc]

    if not stripe_session_obj:
        print("No matching session object found.")
        return

    completed_with_session_object(stripe_session_obj, event_data)


def completed_with_session_object(stripe_session_obj: StripeCheckoutSession, event_data: stripe.checkout.Session) -> None:
    # Fetch current active subscriptions based on the owner (user or organization)
    user_current_plans = UserSubscription.filter_by_owner(owner=stripe_session_obj.owner).filter(end_date__isnull=True)

    # Get plan ID from metadata
    stripe_plan_id = event_data.metadata.get("dj_subscription_plan_id", None) if event_data.metadata else None
    if not stripe_plan_id:
        print("No subscription plan ID found in metadata.")
        return

    # Cancel existing subscriptions except the one in the metadata
    for current_plan in user_current_plans:
        if current_plan.subscription_plan.id != stripe_plan_id:  # Fix: Using `subscription_plan.id`
            stripe.Subscription.modify(current_plan.stripe_subscription_id, cancel_at_period_end=True)  # type: ignore[arg-type]
            current_plan.end_date = timezone_now()
            current_plan.save()

    # Create new subscription if the user doesn't have it
    if not user_current_plans.filter(subscription_plan__id=stripe_plan_id).exists():  # Fix: Using `subscription_plan__id`
        UserSubscription.objects.create(
            owner=stripe_session_obj.owner,
            subscription_plan_id=stripe_plan_id,
            stripe_subscription_id=event_data.subscription,
        )

    # Expire the checkout session
    stripe.checkout.Session.expire(stripe_session_obj.stripe_session_id)  # type: ignore[arg-type]
    stripe_session_obj.delete()



from django.db import connection, OperationalError
from django.http import HttpRequest, HttpResponse
from login_required import login_not_required


@login_not_required
def ping(request: HttpRequest) -> HttpResponse:
    return HttpResponse("pong")


@login_not_required
def healthcheck(request: HttpRequest) -> HttpResponse:
    try:
        connection.ensure_connection()
        return HttpResponse(status=200, content="All operations are up and running!")
    except OperationalError:
        return HttpResponse(status=503, content="Service Unavailable")



from textwrap import dedent

from login_required import login_not_required

from core.service.boto3.handler import BOTO3_HANDLER
from core.types.requests import WebRequest

from django.http import HttpResponse

from core.utils.settings_helper import send_email


@login_not_required
def join_waitlist_endpoint(request: WebRequest):
    email_address = request.POST.get("email", "")
    name = request.POST.get("name", "")

    if not email_address:
        return HttpResponse(status=400)

    if not BOTO3_HANDLER.initiated:
        return HttpResponse(status=500)

    BOTO3_HANDLER.dynamodb_client.put_item(TableName="myfinances-emails", Item={"email": {"S": email_address}, "name": {"S": name}})

    content = """
        <div class='text-success'>
            Successfully registered! Expect some discounts and updates as we progress in our journey :)
        </div>
    """

    send_email(
        destination=email_address,
        subject="Welcome aboard",
        content=dedent(
            f"""
                Thank you for joining our waitlist!

                We're excited to have you on board and will be in touch with more updates as we progress in our journey.

                Stay tuned for discounts, updates and personal direct emails from our founder!

                Best regards,
                The MyFinances Team
            """
        ).strip(),
    )

    return HttpResponse(status=200, content=dedent(content).strip())



import subprocess

from django.core.management import BaseCommand
from django.utils.termcolors import colorize


class Command(BaseCommand):
    help = "Run linters"
    requires_system_checks = []
    requires_migrations_checks = False

    def add_arguments(self, parser):
        parser.add_argument("action", type=str, nargs="?", help="djlint or black")  # parser.add_argument("-f", type=str, dest="flag",

    def handle(self, *args, **kwargs):
        if kwargs["action"] == "djlint":
            djlint()
        elif kwargs["action"] == "black":
            black()
        else:
            self.stdout.write(colorize("Linting with: BLACK FORMATTER", fg="green", opts=("bold",)))
            black()
            self.stdout.write(colorize("Linting with: DJLINT", fg="green", opts=("bold",)))
            djlint()


def djlint():
    subprocess.run(["djlint", "./frontend/templates/", "--reformat"])


def black():
    subprocess.run(["black", "./"])



import logging
from abc import abstractmethod, ABC

from django.http import HttpResponse
from django.shortcuts import render

from core.types.requests import WebRequest

modal_registry = {}


class Modal(ABC):
    modal_name = None  # Override in subclasses
    template_name = None  # Override in subclasses
    context = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if not cls.modal_name:
            raise ValueError(f"Modal class {cls.__name__} must define a `modal_name`.")
        if cls.modal_name in modal_registry:
            logging.info(f"Modal {cls.modal_name} is being overridden by {cls.__name__}.", UserWarning)
        modal_registry[cls.modal_name] = cls

    def get_context(self, request) -> dict:
        """Aggregate context data from classes that explicitly define a `get_context` method."""
        context = {}

        for base in self.__class__.mro():
            # Ensure the base has a distinct `get_context` method to avoid infinite recursion
            if base is Modal:
                continue
            if "get_context" in base.__dict__:
                # logging.debug(f"Gathering context from: {base.__name__}")
                base_context = base.get_context(self, request)
                context.update(base_context)
        return context

    def get(self, request: WebRequest):
        """Populate the context with all inherited context mixins, then return the response."""
        context = self.get_context(request)
        return render(request, self.template_name or f"modals/{self.modal_name}.html", context)

    def get_template_name(self):
        return self.template_name or f"modals/{self.modal_name}.html"

    class _Response:
        def __init__(self, modal_instance, request, context=None, template=None):
            self.modal_instance = modal_instance
            self.request = request
            self.context = context or modal_instance.get_context(request)
            self.template = template or modal_instance.get_template_name()

        def render(self) -> HttpResponse:
            """Return the rendered response."""
            return render(self.request, self.template, self.context)

    def Response(self, request, context=None, template=None) -> HttpResponse:
        """Shortcut to create and render a response."""
        return self._Response(self, request, context, template).render()


def get_modal(name):
    return modal_registry.get(name)


# EXAMPLE MODAL

# class LogoutModal(Modal):
#     modal_name = 'logout'
#
#     def get(self, request: WebRequest, *args, **kwargs):
#         context = {'message': 'Are you sure you want to log out?'}
#         return self.Response(request, context)



# Generated by Django 5.1.4 on 2024-12-31 20:35

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Client",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("active", models.BooleanField(default=True)),
                ("name", models.CharField(max_length=64)),
                ("phone_number", models.CharField(blank=True, max_length=100, null=True)),
                ("email", models.EmailField(blank=True, max_length=254, null=True)),
                ("email_verified", models.BooleanField(default=False)),
                ("company", models.CharField(blank=True, max_length=100, null=True)),
                ("contact_method", models.CharField(blank=True, max_length=100, null=True)),
                ("is_representative", models.BooleanField(default=False)),
                ("address", models.TextField(blank=True, max_length=100, null=True)),
                ("city", models.CharField(blank=True, max_length=100, null=True)),
                ("country", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "organization",
                    models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="core.organization"),
                ),
                (
                    "user",
                    models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                "abstract": False,
                "constraints": [
                    models.CheckConstraint(
                        condition=models.Q(
                            models.Q(("organization__isnull", False), ("user__isnull", True)),
                            models.Q(("organization__isnull", True), ("user__isnull", False)),
                            _connector="OR",
                        ),
                        name="core_client_check_user_or_organization",
                    )
                ],
            },
        ),
    ]



import logging
from core.api.public import APIAuthToken
from rest_framework.decorators import api_view

from core.service.asyn_tasks.tasks import Task
from core.api.public.helpers.response import APIResponse


@api_view(["POST"])
def webhook_task_queue_handler_view_endpoint(request):
    token: APIAuthToken | None = request.auth

    if not token:
        return APIResponse(False, {"status": "error", "message": "No token found"}, status=500)

    if not token.administrator_service_type == token.AdministratorServiceTypes.AWS_WEBHOOK_CALLBACK:
        return APIResponse(False, {"status": "error", "message": "Invalid API key for this service"}, status=500)

    try:
        data: dict = request.data
        func_name: str = data.get("func_name")
        args: list = data.get("args", [])
        kwargs: dict = data.get("kwargs", {})

        print(f"Function Name: {func_name}")
        print(f"Arguments: {args}")
        print(f"Keyword Arguments: {kwargs}")

        # Validate function name
        if not func_name:
            raise ValueError("Function name is required.")

        # Create an instance of Task
        task_helper = Task()

        # Attempt to execute the function
        result = task_helper.execute_now(func_name, *args, **kwargs)

        # Handle the result (e.g., store it or log it)
        print(f"Webhook executed: {func_name} with result: {result}")

        return APIResponse(True, {"status": "success", "result": result})

    except Exception as e:
        logging.error(f"Error executing webhook task: {str(e)}")
        return APIResponse(False, {"status": "error", "message": "An internal error has occurred."}, status=500)



from django.db.models import QuerySet

from core.models import Organization
from core.types.requests import WebRequest


def get_all_users_teams(request: WebRequest) -> QuerySet[Organization]:
    return request.user.teams_joined.all() | request.user.teams_leader_of.all()



from django.core.paginator import Paginator, Page
from django.db.models import Q, QuerySet
from django.http import HttpResponse
from django.shortcuts import render, redirect
from django_ratelimit.core import is_ratelimited

from core.decorators import web_require_scopes
from core.models import EmailSendStatus
from core.types.htmx import HtmxHttpRequest


@web_require_scopes("emails:read", True, True)
def fetch_all_emails(request: HtmxHttpRequest):
    if is_ratelimited(request, group="fetch_all_emails", key="user", rate="2/4s", increment=True) or is_ratelimited(
        request,
        group="fetch_all_emails",
        key="user",
        rate="5/10s",
        increment=True or is_ratelimited(request, group="fetch_all_emails", key="user", rate="20/2m", increment=True),
    ):
        return HttpResponse(status=429)
    context = {}
    if not request.htmx:
        return redirect("quotas")

    search_text = request.GET.get("search")
    page_num = request.GET.get("page")

    if request.user.logged_in_as_team:
        results: QuerySet[EmailSendStatus] = EmailSendStatus.objects.filter(organization=request.user.logged_in_as_team)
    else:
        results = EmailSendStatus.objects.filter(user=request.user)

    if search_text:
        results = results.filter(Q(recipient__icontains=search_text))

    result: Page | QuerySet = results.order_by("-id")

    paginator = Paginator(result, 8)
    result = paginator.get_page(page_num)

    context.update({"emails": result})
    return render(request, "pages/emails/_fetch_body.html", context)



from textwrap import dedent


def recurring_invoices_invoice_created_default_email_template() -> str:
    return dedent(
        """
        Hi $first_name,

        Your invoice #$invoice_id is now available and is due by $due_date. Please make your payment at your earliest convenience.

        Balance Due: $currency_symbol$amount_due $currency
        View or Pay Online: $invoice_link
        If you are paying by standing order, no further action is required. Should you have any questions or concerns, feel free to reach out to us.

        Thank you for your prompt attention to this matter.

        Best regards,
        $company_name
    """
    ).strip()


def recurring_invoices_invoice_overdue_default_email_template() -> str:
    return dedent(
        """
    Hi $first_name,

    We wanted to remind you that invoice #$invoice_id is now overdue. Please arrange payment as soon as possible to ensure there’s no interruption in your service. If you’ve already made the payment, kindly disregard this message—our apologies for any confusion.

    Here are the details for your convenience:

    Balance Due: $currency_symbol$amount_due $currency
    Due Date: $due_date

    If you have any questions or concerns, we’re happy to help. Please don’t hesitate to reach out.

    Thank you for your prompt attention to this matter.

    Warm regards,
    $company_name
    """
    ).strip()


def recurring_invoices_invoice_cancelled_default_email_template() -> str:
    return dedent(
        """
    Hi $first_name,

    The invoice #$invoice_id has been cancelled. You do not have to pay the invoice.

    If you have any questions or concerns, please feel free to contact us.

    Many thanks,
    $company_name
    """
    ).strip()


def email_footer() -> str:
    return (
        "\n\n"
        + dedent(
            """
        Note: This is an automated email sent by MyFinances on behalf of '$company_name'.

        If you believe this email is spam or fraudulent, please do not pay the invoice and report it to us immediately at report@myfinances.cloud.
        Once reported, we will open a case for investigation. In some cases, eligible reports may qualify for a reward, determined on a case-by-case basis.
        """
        ).strip()
    )



from logging import exception
from typing import TypedDict

from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_POST
from django_ratelimit.core import is_ratelimited
from mypy_boto3_sesv2.type_defs import GetMessageInsightsResponseTypeDef, InsightsEventTypeDef

from core.decorators import htmx_only, feature_flag_check, web_require_scopes
from core.utils.settings_helper import EMAIL_CLIENT
from core.models import EmailSendStatus
from core.types.htmx import HtmxHttpRequest


@require_POST
@htmx_only("emails:dashboard")
@feature_flag_check("areUserEmailsAllowed", status=True, api=True, htmx=True)
@web_require_scopes("emails:read", True, True)
def get_status_view(request: HtmxHttpRequest, status_id: str) -> HttpResponse:
    try:
        if request.user.logged_in_as_team:
            EMAIL_STATUS = EmailSendStatus.objects.get(organization=request.user.logged_in_as_team, id=status_id)
        else:
            EMAIL_STATUS = EmailSendStatus.objects.get(user=request.user, id=status_id)
    except EmailSendStatus.DoesNotExist:
        messages.error(request, "Status not found")
        return render(request, "base/toast.html")

    message_insight = get_message_insights(message_id=EMAIL_STATUS.aws_message_id)  # type: ignore[arg-type]

    if isinstance(message_insight, str):
        messages.error(request, message_insight)
        return render(request, "base/toast.html", {"autohide": False})

    important_info = get_important_info_from_response(message_insight)

    EMAIL_STATUS.status = important_info["status"]
    EMAIL_STATUS.updated_status_at = important_info["most_recent_event"]["Timestamp"]
    EMAIL_STATUS.save()

    messages.success(request, f"Status updated to {important_info['status']}")
    return render(request, "base/toast.html", {"autohide": False})


@require_POST
@htmx_only("emails:dashboard")
@feature_flag_check("areUserEmailsAllowed", status=True, api=True, htmx=True)
def refresh_all_statuses_view(request: HtmxHttpRequest) -> HttpResponse:
    if is_ratelimited(request, group="email-refresh_all_statuses", key="user", rate="5/10m", increment=True) or is_ratelimited(
        request, group="email-refresh_all_statuses", key="user", rate="1/m", increment=True
    ):
        messages.error(request, "Woah, slow down! Refreshing the statuses takes a while, give us a break!")
        return render(request, "base/toast.html")
    if request.user.logged_in_as_team:
        ALL_STATUSES = EmailSendStatus.objects.filter(organization=request.user.logged_in_as_team)
    else:
        ALL_STATUSES = EmailSendStatus.objects.filter(user=request.user)

    for status in ALL_STATUSES:
        response = get_message_insights(message_id=status.aws_message_id)  # type: ignore[arg-type]

        if isinstance(response, str):
            messages.error(request, response)
            continue

        important_info = get_important_info_from_response(response)

        status.status = important_info["status"]
        status.updated_status_at = important_info["most_recent_event"]["Timestamp"]

    ALL_STATUSES.bulk_update(ALL_STATUSES, fields=["status", "updated_status_at", "updated_at"])

    messages.success(request, "All statuses have been refreshed")
    http_response = HttpResponse(status=200)
    http_response["HX-Refresh"] = "true"
    return http_response


class ImportantInfo(TypedDict):
    most_recent_event: InsightsEventTypeDef
    status: str


def get_important_info_from_response(response: GetMessageInsightsResponseTypeDef) -> ImportantInfo:
    return {"most_recent_event": (most_recent_event := response["Insights"][0]["Events"][0]), "status": most_recent_event["Type"].lower()}


def get_message_insights(message_id: str) -> GetMessageInsightsResponseTypeDef | str:
    try:
        response = EMAIL_CLIENT.get_message_insights(MessageId=message_id)
        return response
    except EMAIL_CLIENT.exceptions.NotFoundException:
        return "A message was not found with this ID. Maybe wait for it to process"
    except EMAIL_CLIENT.exceptions.BadRequestException:
        return "Something went wrong when trying to fetch the email with this ID"
    except Exception as err:
        exception(err)
        return "Something went wrong when trying to fetch the email with this ID"



from typing import Any

from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from django_htmx.middleware import HtmxDetails

from core.models import User, Organization


class WebRequest(HttpRequest):
    user: User
    team: Organization | None
    team_id: int | None
    actor: User | Organization

    users_subscription: Any | None

    htmx: HtmxDetails
    no_retarget: bool | None



from core.models import User, Organization, TeamMemberPermission
from core.service.permissions.scopes import validate_scopes
from core.utils.dataclasses import BaseServiceResponse


class EditMemberPermissionsServiceResponse(BaseServiceResponse[None]):
    response: None = None


def edit_member_permissions(receiver: User, team: Organization | None, permissions: list) -> EditMemberPermissionsServiceResponse:
    if not validate_receiver(receiver, team):
        return EditMemberPermissionsServiceResponse(error_message="Invalid key name")

    if (scopes_response := validate_scopes(permissions)).failed:
        return EditMemberPermissionsServiceResponse(error_message=scopes_response.error)

    if not team:
        return EditMemberPermissionsServiceResponse(error_message="Invalid team, something went wrong")

    user_team_perms: TeamMemberPermission | None = team.permissions.filter(user=receiver).first()

    if not user_team_perms:
        team.permissions.add(TeamMemberPermission.objects.create(user=receiver, team=team, scopes=permissions))
    else:
        user_team_perms.scopes = permissions
        user_team_perms.save()

    return EditMemberPermissionsServiceResponse(True)


def validate_receiver(receiver: User | None, team: Organization | None) -> bool:
    """
    Make sure receiver is in team and not already owner
    """

    if not receiver:
        return False

    if not team:
        return False

    if not team.members.filter(id=receiver.id).first():
        return False

    if not team.leader == receiver:
        return True
    return False



from core.api.public.permissions import SCOPE_DESCRIPTIONS, SCOPES
from core.types.requests import WebRequest
from core.utils.dataclasses import BaseServiceResponse


class PermissionScopesServiceResponse(BaseServiceResponse[None]):
    response: None = None


def get_permissions_from_request(request: WebRequest) -> list:
    scopes = [
        f"{group}:{perm}"
        for group, items in SCOPE_DESCRIPTIONS.items()
        if (perm := request.POST.get(f"permission_{group}")) in items["options"]
    ]

    scopes.extend(
        f"{group}:read"
        for group, items in SCOPE_DESCRIPTIONS.items()
        if request.POST.get(f"permission_{group}") == "write" and f"{group}:read" in SCOPES
    )

    return scopes


def validate_scopes(permissions: list[str]) -> PermissionScopesServiceResponse:
    """
    Validate permissions are valid
    """
    if not permissions:
        return PermissionScopesServiceResponse(True)

    invalid_permissions: list[str] = [permission for permission in permissions if permission not in SCOPES]

    if invalid_permissions:
        return PermissionScopesServiceResponse(False, error_message=f"Invalid permissions: {', '.join(invalid_permissions)}")

    return PermissionScopesServiceResponse(True)



from typing import Callable

from django.conf import settings

from core.service.maintenance.expire.run import expire_models_task

CORE_MAINTENANCE_TASKS: list[Callable] = [expire_models_task]


def get_maintenance_tasks() -> list[Callable]:
    return getattr(settings, "MAINTENANCE_FUNCTIONS", CORE_MAINTENANCE_TASKS)


def execute_maintenance_tasks(tasks: list[Callable]) -> str:
    output = ""
    for i, task in enumerate(tasks):
        output += f"\n{task()}" if i != 0 else task()
    return output



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from core.decorators import web_require_scopes
from core.models import User
from core.service.permissions.scopes import get_permissions_from_request
from core.service.teams.permissions import edit_member_permissions
from core.types.requests import WebRequest


@require_http_methods(["POST"])
@web_require_scopes("team_permissions:write")
def edit_user_permissions_endpoint(request: WebRequest) -> HttpResponse:
    permissions: list = get_permissions_from_request(request)
    user_id = request.POST.get("user_id")

    receiver: User | None = User.objects.filter(id=user_id).first()

    if not receiver:
        messages.error(request, "Invalid user")
        return render(request, "base/toast.html")

    edit_response = edit_member_permissions(receiver, request.user.logged_in_as_team, permissions)

    if edit_response.success:
        messages.success(request, "User permissions saved successfully")
    else:
        messages.error(request, edit_response.error)
    return render(request, "base/toast.html")



from django.http import HttpResponse
from django.shortcuts import render

from core.decorators import superuser_only
from core.models import QuotaIncreaseRequest, QuotaLimit
from core.types.htmx import HtmxHttpRequest


def quotas_page(request: HtmxHttpRequest) -> HttpResponse:
    groups = list(QuotaLimit.objects.values_list("slug", flat=True).distinct())

    quotas = {q.split("-")[0] for q in groups if q.split("-")}

    return render(
        request,
        "pages/quotas/dashboard.html",
        {"quotas": quotas},
    )


def quotas_list(request: HtmxHttpRequest, group: str) -> HttpResponse:
    return render(request, "pages/quotas/list.html", {"group": group})


@superuser_only
def view_quota_increase_requests(request: HtmxHttpRequest) -> HttpResponse:
    requests = QuotaIncreaseRequest.objects.filter(status="pending").order_by("-created_at")
    return render(request, "pages/quotas/view_requests.html", {"requests": requests})



from django.urls import path
from . import notifications, breadcrumbs
from ...views.modals.open import open_modal_endpoint

# from . import modal

urlpatterns = [
    path(
        "modals/<str:modal_name>/retrieve",
        open_modal_endpoint,
        name="modal retrieve",
    ),
    path(
        "notifications/get",
        notifications.get_notification_html,
        name="notifications get",
    ),
    path("notifications/get_count", notifications.get_notification_count_html, name="notifications get count"),
    path(
        "notifications/delete/<int:id>",
        notifications.delete_notification,
        name="notifications delete",
    ),
    path("breadcrumbs/refetch/", breadcrumbs.update_breadcrumbs_endpoint, name="breadcrumbs refetch"),
]

app_name = "base"



from django.urls import path

from . import change_name, profile_picture, preferences
from .api_keys import generate_api_key_endpoint, revoke_api_key_endpoint

urlpatterns = [
    path(
        "account_preferences/",
        preferences.update_account_preferences,
        name="account_preferences",
    ),
    path(
        "change_name/",
        change_name.change_account_name,
        name="change_name",
    ),
    path("profile_picture/", profile_picture.change_profile_picture_endpoint, name="update profile picture"),
    path("api_keys/generate/", generate_api_key_endpoint, name="api_keys generate"),
    path("api_keys/revoke/<str:key_id>/", revoke_api_key_endpoint, name="api_keys revoke"),
]

app_name = "settings"



from django.contrib import messages
from django.shortcuts import redirect, render
from django.urls import resolve

from core.types.requests import WebRequest
from billing.billing_settings import NO_SUBSCRIPTION_PLAN_DENY_VIEW_NAMES
from billing.models import UserSubscription


# middleware to check if user is subscribed to a plan yet


class CheckUserSubScriptionMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: WebRequest):
        if not request.user.is_authenticated:
            return self.get_response(request)

        if request.team:
            # todo: handle organization billing
            return self.get_response(request)

        subscription: UserSubscription | None = (
            UserSubscription.filter_by_owner(request.actor).filter(end_date__isnull=True).prefetch_related("subscription_plan").first()
        )
        request.users_subscription = subscription

        resolver_match = resolve(request.path_info)

        view_name = resolver_match.view_name

        if view_name not in NO_SUBSCRIPTION_PLAN_DENY_VIEW_NAMES:
            return self.get_response(request)

        if not subscription:
            print("[BILLING] [MIDDLEWARE] User doesn't have an active subscription.")
            messages.warning(
                request,
                """
                You currently are not subscribed to a plan. If you think this is a mistake scroll down and
                press "Refetch" or contact support at
                <a href="mailto:support@strelix.org" class="link link-primary font-extrabold">support@strelix.org</a>.""",
            )

            if request.htmx:
                return render(request, "base/toast.html", {"autohide": False})
            return redirect("billing:dashboard")
        return self.get_response(request)



from datetime import timedelta
from typing import Type, Optional, List

from django.conf import settings
from django.db import models
from django.db.models import QuerySet

from django.apps import apps

from django.utils import timezone

"""
Every model MUST have the field "expires" as:

expires = models.DateTimeField(null=True, blank=True)
"""


def expire_models_task():
    return expire_and_cleanup_objects(getattr(settings, "EXPIRY_MODELS", None))


def expire_and_cleanup_objects(model_list: Optional[List[str]] = None) -> str:
    if model_list is None:
        model_list = ["core.TeamInvitation", "core.PasswordSecret"]

    deactivated_items: int = 0
    deleted_items: int = 0

    now = timezone.now()

    for model in model_list:
        app_label, model_name = model.split(".")

        model_cls = apps.get_model(app_label=app_label, model_name=model_name)

        # Delete objects that have been inactive and expired for more than 14 days
        over_14_days_expired = model_cls.all_objects.filter(expires__lte=now - timedelta(days=14))  # type: ignore[attr-defined]
        deleted_items += over_14_days_expired.count()
        over_14_days_expired.delete()

        # Deactivate expired items that got missed
        to_deactivate: QuerySet[models.Model] = model_cls.all_objects.filter(expires__lte=now, active=True)  # type: ignore[attr-defined]

        deactivated_items += to_deactivate.count()
        to_deactivate.update(active=False)

    return f"Deactivated {deactivated_items} objects and deleted {deleted_items} objects."



from textwrap import dedent

from django.urls import reverse
from django.utils.crypto import get_random_string

from core.models import User, Organization, TeamMemberPermission
from core.utils.dataclasses import BaseServiceResponse
from core.utils.settings_helper import send_email


class CreateUserServiceResponse(BaseServiceResponse[User]): ...


def create_user_service(
    request, email: str, team: Organization, first_name: str, last_name: str, permissions: list[str]
) -> CreateUserServiceResponse:

    if not first_name:
        return CreateUserServiceResponse(error_message="Please enter a valid first name")

    if not email:
        return CreateUserServiceResponse(error_message="Please enter a valid user email")

    if User.objects.filter(email=email).exists():
        return CreateUserServiceResponse(error_message="This user already exists, invite them instead!")

    temporary_password = get_random_string(length=8)

    user: User = User.objects.create_user(email=email, first_name=first_name, last_name=last_name, username=email)
    user.set_password(temporary_password)
    user.awaiting_email_verification = False
    user.require_change_password = True
    user.save()

    send_email(
        destination=email,
        subject="You have been invited to join an organization",
        content=dedent(
            f"""
                Hi {user.first_name or "User"},

                You have been invited by {request.user.email} to join the organization '{team.name}'.

                Your account email is: {email}
                Your temporary password is: {temporary_password}

                We suggest that you change your password as soon as you login, however no other user including the organization have
                access to this password.

                Upon login, you will be added to the \"{team.name}\" organization. However, if required, you may leave at any point.

                Login to your new account using this link:
                {request.build_absolute_uri(reverse("core:auth:login"))}

                Didn't give permission to be added to this organization? You can safely ignore the email, no actions can be done on
                behalf of you without your permission.
            """
        ),
    )

    team.members.add(user)

    TeamMemberPermission.objects.create(user=user, team=team, scopes=permissions)

    return CreateUserServiceResponse(True, response=user)
